import click
from colorama import Fore, Style

from vantage6.common import error, info
from vantage6.common.globals import InstanceType

from vantage6.cli.common.stop import helm_uninstall, stop_port_forward
from vantage6.cli.common.utils import (
    find_running_service_names,
    select_context_and_namespace,
    select_running_service,
)
from vantage6.cli.context import get_context
from vantage6.cli.globals import DEFAULT_SERVER_SYSTEM_FOLDERS


@click.command()
@click.option("-n", "--name", default=None, help="Configuration name")
@click.option("--context", default=None, help="Kubernetes context to use")
@click.option("--namespace", default=None, help="Kubernetes namespace to use")
@click.option(
    "--system",
    "system_folders",
    flag_value=True,
    default=DEFAULT_SERVER_SYSTEM_FOLDERS,
    help="Search for configuration in system folders instead of user folders. "
    "This is the default.",
)
@click.option(
    "--user",
    "system_folders",
    flag_value=False,
    help="Search for configuration in the user folders instead of system folders.",
)
@click.option("--all", "all_stores", flag_value=True, help="Stop all algorithm stores")
def cli_algo_store_stop(
    name: str,
    context: str,
    namespace: str,
    system_folders: bool,
    all_stores: bool,
):
    """
    Stop one or all running algorithm store(s).
    """
    context, namespace = select_context_and_namespace(
        context=context,
        namespace=namespace,
    )

    running_stores = find_running_service_names(
        instance_type=InstanceType.ALGORITHM_STORE,
        only_system_folders=system_folders,
        only_user_folders=not system_folders,
        context=context,
        namespace=namespace,
    )

    if not running_stores:
        error("No running algorithm stores found.")
        return

    if all_stores:
        for store in running_stores:
            _stop_store(store["name"], namespace, context)
    else:
        if not name:
            store_name = select_running_service(
                running_stores, InstanceType.ALGORITHM_STORE
            )
        else:
            ctx = get_context(InstanceType.ALGORITHM_STORE, name, system_folders)
            store_name = ctx.helm_release_name

        if store_name in running_stores:
            _stop_store(store_name, namespace, context)
            info(f"Stopped the {Fore.GREEN}{store_name}{Style.RESET_ALL} store.")
        else:
            error(f"{Fore.RED}{name}{Style.RESET_ALL} is not running?!")


def _stop_store(store_name: str, namespace: str, context: str) -> None:
    info(f"Stopping store {store_name}...")

    # uninstall the helm release
    helm_uninstall(
        release_name=store_name,
        context=context,
        namespace=namespace,
    )

    stop_port_forward(
        service_name=f"{store_name}-vantage6-algorithm-store-service",
    )

    info(f"Store {store_name} stopped successfully.")



import click

from vantage6.common import info
from vantage6.cli.common.utils import attach_logs


@click.command()
def cli_algo_store_attach() -> None:
    """
    Show the store logs in the current console.
    """
    info("Attaching to store logs...")
    attach_logs("app=store", "component=store-server")



import os
from pathlib import Path
from typing import Any

import questionary as q

from vantage6.common import error, info, warning
from vantage6.common.client.node_client import NodeClient
from vantage6.common.context import AppContext
from vantage6.common.globals import (
    DATABASE_TYPES,
    DEFAULT_API_PATH,
    InstanceType,
    NodePolicy,
    Ports,
    RequiredNodeEnvVars,
)

from vantage6.cli.config import CliConfig
from vantage6.cli.configuration_manager import (
    AlgorithmStoreConfigurationManager,
    NodeConfigurationManager,
    ServerConfigurationManager,
)
from vantage6.cli.context import select_context_class


def node_configuration_questionaire(dirs: dict, instance_name: str) -> dict:
    """
    Questionary to generate a config file for the node instance.

    Parameters
    ----------
    dirs : dict
        Dictionary with the directories of the node instance.
    instance_name : str
        Name of the node instance.

    Returns
    -------
    dict
        Dictionary with the new node configuration
    """
    config = q.unsafe_prompt(
        [
            {"type": "text", "name": "api_key", "message": "Enter given api-key:"},
            {
                "type": "text",
                "name": "server_url",
                "message": "The base-URL of the server:",
                "default": "http://localhost",
            },
        ]
    )
    # remove trailing slash from server_url if entered by user
    config["server_url"] = config["server_url"].rstrip("/")

    # set default port to the https port if server_url is https
    default_port = (
        str(Ports.HTTPS)
        if config["server_url"].startswith("https")
        else str(Ports.DEV_SERVER)
    )

    config = config | q.unsafe_prompt(
        [
            {
                "type": "text",
                "name": "port",
                "message": "Enter port to which the server listens:",
                "default": default_port,
            },
            {
                "type": "text",
                "name": "api_path",
                "message": "Path of the api:",
                "default": "/api",
            },
            {
                "type": "text",
                "name": "task_dir",
                "message": "Task directory path:",
                "default": str(dirs["data"]),
            },
        ]
    )

    config["databases"] = list()
    while q.confirm("Do you want to add a database?").unsafe_ask():
        db_label = q.unsafe_prompt(
            [
                {
                    "type": "text",
                    "name": "label",
                    "message": "Enter unique label for the database:",
                    "default": "default",
                }
            ]
        )
        db_path = q.unsafe_prompt(
            [{"type": "text", "name": "uri", "message": "Database URI:"}]
        )
        db_type = q.select("Database type:", choices=DATABASE_TYPES).unsafe_ask()

        config["databases"].append(
            {"label": db_label.get("label"), "uri": db_path.get("uri"), "type": db_type}
        )

    is_add_vpn = q.confirm(
        "Do you want to connect to a VPN server?", default=False
    ).unsafe_ask()
    if is_add_vpn:
        config["vpn_subnet"] = q.text(
            message="Subnet of the VPN server you want to connect to:",
            default="10.76.0.0/16",
        ).unsafe_ask()

    is_policies = q.confirm(
        "Do you want to limit the algorithms allowed to run on your node? This "
        "should always be done for production scenarios.",
        default=True,
    ).unsafe_ask()
    policies = {}
    if is_policies:
        info(
            "You can limit the algorithms that can run on your node in two ways: by "
            "allowing specific algorithms or by allowing all algorithms in a given "
            "algorithm store."
        )
        ask_single_algorithms = q.confirm(
            "Do you want to enter a list of allowed algorithms?"
        ).unsafe_ask()
        if ask_single_algorithms:
            policies[NodePolicy.ALLOWED_ALGORITHMS.value] = _get_allowed_algorithms()
        ask_algorithm_stores = q.confirm(
            "Do you want to allow algorithms from specific algorithm stores?"
        ).unsafe_ask()
        if ask_algorithm_stores:
            policies[NodePolicy.ALLOWED_ALGORITHM_STORES.value] = (
                _get_allowed_algorithm_stores()
            )
        if ask_single_algorithms and ask_algorithm_stores:
            require_both_whitelists = q.confirm(
                "Do you want to allow only algorithms that are both in the list of "
                "allowed algorithms *AND* are part of one of the allowed algorithm "
                "stores? If not, algorithms will be allowed if they are in either the "
                "list of allowed algorithms or one of the allowed algorithm stores.",
                default=True,
            ).unsafe_ask()
            policies["allow_either_whitelist_or_store"] = not require_both_whitelists
    if policies:
        config["policies"] = policies

    res = q.select(
        "Which level of logging would you like?",
        choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL", "NOTSET"],
    ).unsafe_ask()

    config["logging"] = {
        "level": res,
        "use_console": True,
        "backup_count": 5,
        "max_size": 1024,
        "format": "%(asctime)s - %(name)-14s - %(levelname)-8s - %(message)s",
        "datefmt": "%Y-%m-%d %H:%M:%S",
        "loggers": [
            {"name": "urllib3", "level": "warning"},
            {"name": "requests", "level": "warning"},
            {"name": "engineio.client", "level": "warning"},
            {"name": "docker.utils.config", "level": "warning"},
            {"name": "docker.auth", "level": "warning"},
        ],
    }

    # Check if we can login to the server to retrieve collaboration settings
    client = NodeClient(
        instance_name,
        config["api_key"],
        server_url=f"{config['server_url']}:{config['port']}{config['api_path']}",
        auth_url=os.environ.get(RequiredNodeEnvVars.KEYCLOAK_URL.value),
    )
    try:
        client.authenticate()
    except Exception as e:
        error(f"Could not authenticate with server: {e}")
        error("Please check (1) your API key and (2) if your server is online")
        warning(
            "If you continue, you should provide your collaboration settings manually."
        )
        if q.confirm("Do you want to abort?", default=True).unsafe_ask():
            exit(0)

    if client.whoami is not None:
        encryption = client.is_encrypted_collaboration()
        # TODO when we build collaboration policies, update this to provide
        # the node admin with a list of all policies, and whether or not
        # to accept them
        q.confirm(
            f"Encryption is {'enabled' if encryption else 'disabled'}"
            f" for this collaboration. Accept?",
            default=True,
        ).unsafe_ask()
    else:
        encryption = q.confirm("Enable encryption?", default=True).unsafe_ask()

    private_key = (
        "" if not encryption else q.text("Path to private key file:").unsafe_ask()
    )

    config["encryption"] = {
        "enabled": encryption is True or encryption == "true",
        "private_key": private_key,
    }

    return config


def _get_allowed_algorithms() -> list[str]:
    """
    Prompt the user for the allowed algorithms on their node

    Returns
    -------
    list[str]
        List of allowed algorithms or regular expressions to match them
    """
    info("Below you can add algorithms that are allowed to run on your node.")
    info(
        "You can use regular expressions to match multiple algorithms, or you can "
        "use strings to provide one algorithm at a time."
    )
    info("Examples:")
    info(r"^harbor2\.vantage6\.ai/demo/average$    Allow the demo average algorithm")
    info(
        r"^harbor2\.vantage6\.ai/algorithms/.*   Allow all algorithms from "
        "harbor2.vantage6.ai/algorithms"
    )
    info(
        r"^harbor2\.vantage6\.ai/demo/average@sha256:82becede...$    Allow a "
        "specific hash of average algorithm"
    )
    allowed_algorithms = []
    while True:
        algo = q.text(message="Enter your algorithm expression:").unsafe_ask()
        allowed_algorithms.append(algo)
        if not q.confirm(
            "Do you want to add another algorithm expression?", default=True
        ).unsafe_ask():
            break
    return allowed_algorithms


def _get_allowed_algorithm_stores() -> list[str]:
    """
    Prompt the user for the allowed algorithm stores on their node

    Returns
    -------
    list[str]
        List of allowed algorithm stores
    """
    info("Below you can add algorithm stores that are allowed to run on your node.")
    info(
        "You can use regular expressions to match multiple algorithm stores, or you can"
        " use strings to provide one algorithm store at a time."
    )
    info("Examples:")
    info(
        "https://store.cotopaxi.vantage6.ai    Allow all algorithms from the "
        "community store"
    )
    info(
        r"^https://*\.vantage6\.ai$               Allow all algorithms from any "
        "store hosted on vantage6.ai"
    )
    allowed_algorithm_stores = []
    while True:
        store = q.text(message="Enter the URL of the algorithm store:").unsafe_ask()
        allowed_algorithm_stores.append(store)
        if not q.confirm(
            "Do you want to add another algorithm store?", default=True
        ).unsafe_ask():
            break
    return allowed_algorithm_stores


def _add_common_server_config(
    config: dict, instance_type: InstanceType, instance_name: str
) -> dict:
    """
    Part of the questionaire that is common to all server types (vantage6
    server and algorithm store server).

    Parameters
    ----------
    instance_type : InstanceType
        Type of server instance.
    instance_name : str
        Name of the server instance.

    Returns
    -------
    dict
        Dictionary with new (partial) server configuration
    """
    backend_config = (
        config["server"] if instance_type == InstanceType.SERVER else config["store"]
    )

    backend_config["port"] = q.text(
        "Enter port to which the server listens:",
        default=(
            str(Ports.DEV_SERVER)
            if instance_type == InstanceType.SERVER
            else str(Ports.DEV_ALGO_STORE)
        ),
    ).unsafe_ask()

    res = q.select(
        "Which level of logging would you like?",
        choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
        default="INFO",
    ).unsafe_ask()

    backend_config["logging"] = {
        "level": res,
        "file": f"{instance_name}.log",
        "use_console": True,
        "backup_count": 5,
        "max_size": 1024,
        "format": "%(asctime)s - %(name)-14s - %(levelname)-8s - %(message)s",
        "datefmt": "%Y-%m-%d %H:%M:%S",
        "loggers": [
            {"name": "urllib3", "level": "warning"},
            {"name": "socketIO-client", "level": "warning"},
            {"name": "socketio.server", "level": "warning"},
            {"name": "engineio.server", "level": "warning"},
            {"name": "sqlalchemy.engine", "level": "warning"},
        ],
    }

    backend_config["api_path"] = DEFAULT_API_PATH

    # === Database settings ===
    config["database"]["volumePath"] = q.text(
        "Where is your server database located on the host machine?",
        default=f"{Path.cwd()}/dev/.db/db_pv_server",
    ).unsafe_ask()

    config["database"]["k8sNodeName"] = q.text(
        "What is the name of the k8s node where the databases are running?",
        default="docker-desktop",
    ).unsafe_ask()

    is_production = q.confirm(
        "Do you want to use production settings for this server? If not, the server will "
        "be configured to be more suitable for development or testing purposes.",
        default=True,
    ).unsafe_ask()

    if is_production:
        config = _add_production_server_config(config)

    return config, is_production


def server_configuration_questionaire(instance_name: str) -> dict[str, Any]:
    """
    Kubernetes-specific questionnaire to generate Helm values for server.

    Parameters
    ----------
    instance_name : str
        Name of the server instance.

    Returns
    -------
    dict[str, Any]
        dictionary with Helm values for the server configuration
    """
    # Get active kube namespace
    cli_config = CliConfig()
    kube_namespace = cli_config.get_last_namespace()

    # Initialize config with basic structure
    config = {"server": {}, "database": {}, "ui": {}, "rabbitmq": {}}

    config, is_production = _add_common_server_config(
        config, InstanceType.SERVER, instance_name
    )
    if not is_production:
        config["server"]["jwt"] = {
            "secret": "constant_development_secret`",
        }
        config["server"]["dev"] = {
            "host_uri": "host.docker.internal",
            "store_in_local_cluster": True,
        }

    # TODO v5+ these should be removed, latest should usually be used so question is
    # not needed. However, for now we want to specify alpha/beta images.
    # === Server settings ===
    config["server"]["image"] = q.text(
        "Server Docker image:",
        default="harbor2.vantage6.ai/infrastructure/server:latest",
    ).unsafe_ask()

    # === UI settings ===
    config["ui"]["image"] = q.text(
        "UI Docker image:",
        default="harbor2.vantage6.ai/infrastructure/ui:latest",
    ).unsafe_ask()

    # === Keycloak settings ===
    keycloak_url = f"http://vantage6-auth-keycloak.{kube_namespace}.svc.cluster.local"
    config["server"]["keycloakUrl"] = keycloak_url

    return config


def _add_production_server_config(config: dict) -> dict:
    """
    Add the production server configuration to the config

    Parameters
    ----------
    config : dict
        The config to add the production server configuration to

    Returns
    -------
    dict
        The config with the production server configuration added
    """
    info("For production environments, it is recommended to use an external database.")
    info("Please provide the URI of the external database.")
    info("Example: postgresql://username:password@localhost:5432/vantage6")

    config["database"]["external"] = True
    config["database"]["uri"] = q.text(
        "Database URI:",
        default="postgresql://vantage6:vantage6@localhost:5432/vantage6",
    ).unsafe_ask()

    return config


def algo_store_configuration_questionaire(instance_name: str) -> dict:
    """
    Questionary to generate a config file for the algorithm store server
    instance.

    Parameters
    ----------
    instance_name : str
        Name of the server instance.

    Returns
    -------
    dict
        Dictionary with the new server configuration
    """
    config = {"store": {}, "database": {}}

    config, is_production = _add_common_server_config(
        config, InstanceType.ALGORITHM_STORE, instance_name
    )
    if not is_production:
        config["store"]["dev"] = {
            "host_uri": "host.docker.internal",
            "disable_review": True,
            "review_own_algorithm": True,
        }

    default_v6_server_uri = f"http://localhost:{Ports.DEV_SERVER}{DEFAULT_API_PATH}"
    default_root_username = "admin"

    v6_server_uri = q.text(
        "What is the Vantage6 server linked to the algorithm store? "
        "Provide the link to the server endpoint.",
        default=default_v6_server_uri,
    ).unsafe_ask()

    root_username = q.text(
        "What is the username of the root user?",
        default=default_root_username,
    ).unsafe_ask()

    config["root_user"] = {
        "v6_server_uri": v6_server_uri,
        "username": root_username,
    }

    # ask about openness of the algorithm store
    config["policies"] = {}
    is_open = q.confirm(
        "Do you want to open the algorithm store to the public? This will allow anyone "
        "to view the algorithms, but they cannot modify them.",
        default=False,
    ).unsafe_ask()
    if is_open:
        open_algos_policy = "public"
    else:
        is_open_to_whitelist = q.confirm(
            "Do you want to allow all authenticated users to access "
            "the algorithms in the store? If not allowing this, you will have to assign"
            " the appropriate permissions to each user individually.",
            default=True,
        ).unsafe_ask()
        open_algos_policy = "authenticated" if is_open_to_whitelist else "private"
    config["policies"]["algorithm_view"] = open_algos_policy

    return config


def configuration_wizard(
    type_: InstanceType,
    instance_name: str,
    system_folders: bool,
) -> Path:
    """
    Create a configuration file for a node or server instance.

    Parameters
    ----------
    type_ : InstanceType
        Type of the instance to create a configuration for
    instance_name : str
        Name of the instance
    system_folders : bool
        Whether to use the system folders or not

    Returns
    -------
    Path
        Path to the configuration file
    """
    # for defaults and where to save the config
    dirs = AppContext.instance_folders(type_, instance_name, system_folders)

    # invoke questionaire to create configuration file
    if type_ == InstanceType.NODE:
        conf_manager = NodeConfigurationManager
        config = node_configuration_questionaire(dirs, instance_name)
    elif type_ == InstanceType.SERVER:
        conf_manager = ServerConfigurationManager
        config = server_configuration_questionaire(instance_name)
    else:
        conf_manager = AlgorithmStoreConfigurationManager
        config = algo_store_configuration_questionaire(instance_name)

    # in the case of an environment we need to add it to the current
    # configuration. In the case of application we can simply overwrite this
    # key (although there might be environments present)
    config_file = Path(dirs.get("config")) / (instance_name + ".yaml")

    if Path(config_file).exists():
        config_manager = conf_manager.from_file(config_file)
    else:
        config_manager = conf_manager(instance_name)

    config_manager.put(config)
    config_manager.save(config_file)

    return config_file


def select_configuration_questionaire(type_: InstanceType, system_folders: bool) -> str:
    """
    Ask which configuration the user wants to use. It shows only configurations
    that are in the default folder.

    Parameters
    ----------
    type_ : InstanceType
        Type of the instance to create a configuration for
    system_folders : bool
        Whether to use the system folders or not

    Returns
    -------
    str
        Name of the configuration
    """
    context = select_context_class(type_)
    configs, _ = context.available_configurations(system_folders)

    # each collection (file) can contain multiple configs. (e.g. test,
    # dev)
    choices = []
    for config_collection in configs:
        choices.append(
            q.Choice(title=f"{config_collection.name:25}", value=config_collection.name)
        )

    if not choices:
        raise Exception("No configurations could be found!")

    # pop the question
    try:
        return q.select(
            "Select the configuration you want to use:", choices=choices
        ).unsafe_ask()
    except KeyboardInterrupt:
        error("Aborted by user!")
        exit(1)



import click
import questionary as q
import docker
import os.path
import itertools

from pathlib import Path
from shutil import rmtree

from vantage6.common import (
    error,
    info,
    debug,
)
from vantage6.common.globals import APPNAME


from vantage6.cli.context.node import NodeContext
from vantage6.cli.globals import DEFAULT_NODE_SYSTEM_FOLDERS as N_FOL
from vantage6.cli.utils import check_if_docker_daemon_is_running, remove_file
from vantage6.cli.node.common import select_node, find_running_node_names

# TODO v5+ remove this - just a dummy to prevent import issues from v4 CLI
# from vantage6.common.globals import VPN_CONFIG_FILE
VPN_CONFIG_FILE = "vpn.conf"


@click.command()
@click.option("-n", "--name", default=None, help="Configuration name")
@click.option(
    "--system",
    "system_folders",
    flag_value=True,
    help="Search for configuration in system folders rather than " "user folders",
)
@click.option(
    "--user",
    "system_folders",
    flag_value=False,
    default=N_FOL,
    help="Search for configuration in user folders rather than "
    "system folders. This is the default",
)
@click.option(
    "-f", "--force", type=bool, flag_value=True, help="Don't ask for confirmation"
)
def cli_node_remove(name: str, system_folders: bool, force: bool) -> None:
    """
    Delete a node permanently.

    Remove the configuration file, log file, and docker volumes attached to
    the node.
    """
    # select configuration name if none supplied
    name = select_node(name, system_folders)

    client = docker.from_env()
    check_if_docker_daemon_is_running(client)

    # check if node is still running, otherwise don't allow deleting it
    running_node_names = find_running_node_names(client)

    post_fix = "system" if system_folders else "user"
    node_container_name = f"{APPNAME}-{name}-{post_fix}"
    if node_container_name in running_node_names:
        error(
            f"Node {name} is still running! Please stop the node before " "deleting it."
        )
        exit(1)

    if not force:
        if not q.confirm(
            "This node will be deleted permanently including its "
            "configuration. Are you sure?",
            default=False,
        ).ask():
            info("Node will not be deleted")
            exit(0)

    # create node context
    ctx = NodeContext(name, system_folders=system_folders)

    # remove the docker volume and any temporary volumes
    debug("Deleting docker volumes")
    volumes = client.volumes.list()
    for vol in volumes:
        if vol.name.startswith(ctx.docker_volume_name):  # includes tmp volumes
            info(f"Deleting docker volume {vol.name}")
            vol.remove()
        # remove docker vpn volume
        if vol.name == ctx.docker_vpn_volume_name:
            info(f"Deleting VPN docker volume {vol.name}")
            vol.remove()

    # remove the VPN configuration file
    vpn_config_file = os.path.join(ctx.data_dir, "vpn", VPN_CONFIG_FILE)
    remove_file(vpn_config_file, "VPN configuration")

    # remove the config file
    remove_file(ctx.config_file, "configuration")

    # remove the log file. As this process opens the log file above, the log
    # handlers need to be closed before deleting
    log_dir = Path(ctx.log_file.parent)
    info(f"Removing log file {log_dir}")
    for handler in itertools.chain(ctx.log.handlers, ctx.log.root.handlers):
        handler.close()
    # remove the whole folder with all the log files
    rmtree(log_dir)

    # remove the folder: if it hasn't been started yet this won't exist...
    if Path.exists(ctx.config_dir / name):
        rmtree(ctx.config_dir / name)



import click

from vantage6.common.globals import InstanceType

from vantage6.cli.common.new import new
from vantage6.cli.globals import DEFAULT_SERVER_SYSTEM_FOLDERS


@click.command()
@click.option(
    "-n", "--name", default=None, help="name of the configuration you want to use."
)
@click.option(
    "--system",
    "system_folders",
    flag_value=True,
    help="Use system folders instead of user folders. This is the default",
)
@click.option(
    "--user",
    "system_folders",
    flag_value=False,
    default=DEFAULT_SERVER_SYSTEM_FOLDERS,
    help="Use user folders instead of system folders",
)
@click.option("--context", default=None, help="Kubernetes context to use")
@click.option(
    "--namespace",
    default=None,
    help="Kubernetes namespace to use",
)
def cli_server_new(
    name: str,
    system_folders: bool,
    namespace: str,
    context: str,
) -> None:
    """
    Create a new server configuration.
    """
    new(name, system_folders, namespace, context, InstanceType.SERVER)



from __future__ import annotations

import subprocess

from vantage6.common import error, info, warning

from vantage6.cli.utils import validate_input_cmd_args


def stop_port_forward(service_name: str) -> None:
    """
    Stop the port forwarding process for a given service name.

    Parameters
    ----------
    service_name : str
        The name of the service whose port forwarding process should be terminated.
    """
    # Input validation
    validate_input_cmd_args(service_name, "service name")

    try:
        # Find the process ID (PID) of the port forwarding command
        result = subprocess.run(
            ["pgrep", "-f", f"kubectl port-forward.*{service_name}"],
            check=True,
            text=True,
            capture_output=True,
        )
        pids = result.stdout.strip().splitlines()

        if not pids:
            warning(f"No port forwarding process found for service '{service_name}'.")
            return

        for pid in pids:
            subprocess.run(["kill", "-9", pid], check=True)
            info(
                f"Terminated port forwarding process for service '{service_name}' "
                f"(PID: {pid})"
            )
    except subprocess.CalledProcessError as e:
        error(f"Failed to terminate port forwarding: {e}")


def helm_uninstall(
    release_name: str,
    context: str | None = None,
    namespace: str | None = None,
) -> None:
    """
    Manage the `helm uninstall` command.

    Parameters
    ----------
    release_name : str
        The name of the Helm release to uninstall.
    context : str, optional
        The Kubernetes context to use.
    namespace : str, optional
        The Kubernetes namespace to use.
    """
    # Input validation
    validate_input_cmd_args(release_name, "release name")
    validate_input_cmd_args(context, "context name", allow_none=True)
    validate_input_cmd_args(namespace, "namespace name", allow_none=True)

    # Create the command
    command = ["helm", "uninstall", release_name]

    if context:
        command.extend(["--kube-context", context])

    if namespace:
        command.extend(["--namespace", namespace])

    try:
        subprocess.run(
            command,
            stdout=subprocess.DEVNULL,
            check=True,
        )
        info(f"Successfully uninstalled release '{release_name}'.")
    except subprocess.CalledProcessError as e:
        error(f"Failed to uninstall release '{release_name}': {e.stderr}")
    except FileNotFoundError:
        error(
            "Helm command not found. Please ensure Helm is installed and available in "
            "the PATH."
        )



import unittest
import logging
import os
import contextlib

from unittest.mock import MagicMock, patch
from pathlib import Path
from io import BytesIO, StringIO
from click.testing import CliRunner
from docker.errors import APIError

from vantage6.common.globals import Ports
from vantage6.cli.globals import APPNAME
from vantage6.common import STRING_ENCODING
from vantage6.cli.common.utils import print_log_worker
from vantage6.cli.node.list import cli_node_list
from vantage6.cli.node.new import cli_node_new_configuration
from vantage6.cli.node.files import cli_node_files
from vantage6.cli.node.start import cli_node_start
from vantage6.cli.node.restart import cli_node_restart
from vantage6.cli.node.stop import cli_node_stop
from vantage6.cli.node.attach import cli_node_attach
from vantage6.cli.node.create_private_key import cli_node_create_private_key
from vantage6.cli.node.clean import cli_node_clean
from vantage6.cli.node.common import create_client_and_authenticate


class NodeCLITest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        logging.getLogger("docker.utils.config").setLevel(logging.WARNING)
        return super().setUpClass()

    @patch("docker.DockerClient.ping")
    def test_list_docker_not_running(self, docker_ping):
        """An error is printed when docker is not running"""
        docker_ping.side_effect = Exception("Boom!")

        runner = CliRunner()
        result = runner.invoke(cli_node_list, [])

        # check exit code
        self.assertEqual(result.exit_code, 1)

    @patch("vantage6.cli.context.node.NodeContext.available_configurations")
    @patch("docker.DockerClient.ping")
    @patch("docker.DockerClient.containers")
    def test_list(self, containers, docker_ping, available_configurations):
        """A container list and their current status."""
        # https://docs.python.org/3/library/unittest.mock.html#mock-names-and-the-name-attribute

        # mock that docker-deamon is running
        docker_ping.return_value = True

        # docker deamon returns a list of running node-containers
        container1 = MagicMock()
        container1.name = f"{APPNAME}-iknl-user"
        containers.list.return_value = [container1]

        # returns a list of configurations and failed inports
        def side_effect(system_folders):
            config = MagicMock()
            config.name = "iknl"
            if not system_folders:
                return [[config], []]
            else:
                return [[config], []]

        available_configurations.side_effect = side_effect

        # invoke CLI method
        runner = CliRunner()
        result = runner.invoke(cli_node_list, [])

        # validate exit code
        self.assertEqual(result.exit_code, 0)

        # check printed lines
        self.assertEqual(
            result.output,
            "\nName                     Status          System/User\n"
            "-----------------------------------------------------\n"
            "iknl                     Not running     System \n"
            "iknl                     Running         User   \n"
            "-----------------------------------------------------\n",
        )

    @patch("vantage6.cli.node.new.configuration_wizard")
    @patch("vantage6.cli.node.new.ensure_config_dir_writable")
    @patch("vantage6.cli.node.common.NodeContext")
    def test_new_config(self, context, permissions, wizard):
        """No error produced when creating new configuration."""
        context.config_exists.return_value = False
        permissions.return_value = True
        wizard.return_value = "/some/file/path"

        runner = CliRunner()
        result = runner.invoke(
            cli_node_new_configuration,
            [
                "--name",
                "some-name",
            ],
        )

        # check that info message is produced
        self.assertEqual(result.output[:7], "[info ]")

        # check OK exit code
        self.assertEqual(result.exit_code, 0)

    @patch("vantage6.cli.node.new.configuration_wizard")
    def test_new_config_replace_whitespace_in_name(self, _):
        """Whitespaces are replaced in the name."""

        runner = CliRunner()
        result = runner.invoke(
            cli_node_new_configuration,
            [
                "--name",
                "some name",
            ],
        )

        self.assertEqual(
            result.output[:60],
            "[info ] - Replaced spaces from configuration name: some-name",
        )

    @patch("vantage6.cli.node.new.NodeContext")
    def test_new_config_already_exists(self, context):
        """No duplicate configurations are allowed."""

        context.config_exists.return_value = True

        runner = CliRunner()
        result = runner.invoke(
            cli_node_new_configuration,
            [
                "--name",
                "some-name",
            ],
        )

        # check that error is produced
        self.assertEqual(result.output[:7], "[error]")

        # check non-zero exit code
        self.assertEqual(result.exit_code, 1)

    @patch("vantage6.cli.node.new.ensure_config_dir_writable")
    @patch("vantage6.cli.node.common.NodeContext")
    def test_new_write_permissions(self, context, permissions):
        """User needs write permissions."""

        context.config_exists.return_value = False
        permissions.return_value = False

        runner = CliRunner()
        result = runner.invoke(
            cli_node_new_configuration,
            [
                "--name",
                "some-name",
            ],
        )

        # check that error is produced
        self.assertEqual(result.output[:7], "[error]")

        # check non-zero exit code
        self.assertEqual(result.exit_code, 1)

    @patch("vantage6.cli.node.common.NodeContext")
    @patch("vantage6.cli.node.files.NodeContext")
    @patch("vantage6.cli.node.common.select_configuration_questionaire")
    def test_files(self, select_config, context, common_context):
        """No errors produced when retrieving filepaths."""

        common_context.config_exists.return_value = True
        context.return_value = MagicMock(
            config_file="/file.yaml", log_file="/log.log", data_dir="/dir"
        )
        context.return_value.databases.items.return_value = [["label", "/file.db"]]
        select_config.return_value = "iknl"

        runner = CliRunner()
        result = runner.invoke(cli_node_files, [])

        # we check that no warnings have been produced
        self.assertEqual(result.output[:7], "[info ]")

        # check status code is OK
        self.assertEqual(result.exit_code, 0)

    @patch("vantage6.cli.node.common.NodeContext")
    def test_files_non_existing_config(self, context):
        """An error is produced when a non existing config is used."""

        context.config_exists.return_value = False

        runner = CliRunner()
        result = runner.invoke(cli_node_files, ["--name", "non-existing"])

        # Check that error is produced
        self.assertEqual(result.output[:7], "[error]")

        # check for non zero exit-code
        self.assertNotEqual(result.exit_code, 0)

    @patch("docker.DockerClient.volumes")
    @patch("vantage6.cli.node.start.pull_infra_image")
    @patch("vantage6.cli.common.decorator.get_context")
    @patch("docker.DockerClient.containers")
    @patch("vantage6.cli.node.start.check_docker_running", return_value=True)
    def test_start(self, check_docker, client, context, pull, volumes):
        # client.containers = MagicMock(name="docker.DockerClient.containers")
        client.list.return_value = []
        volume = MagicMock()
        volume.name = "data-vol-name"
        volumes.create.return_value = volume
        context.config_exists.return_value = True

        ctx = MagicMock(
            data_dir=Path("data"),
            log_dir=Path("logs"),
            config_dir=Path("configs"),
            databases=[{"label": "some_label", "uri": "data.csv", "type": "csv"}],
        )

        # cli_node_start() tests for truth value of a set-like object derived
        # from ctx.config.get('node_extra_env', {}). Default MagicMock() will
        # evaluate to True, empty dict to False. False signifies no overwritten
        # env vars, hence no error.
        def config_get_side_effect(key, default=None):
            if key == "node_extra_env":
                return {}
            return MagicMock()

        ctx.config.get.side_effect = config_get_side_effect
        ctx.get_data_file.return_value = "data.csv"
        ctx.name = "some-name"
        context.return_value = ctx

        runner = CliRunner()

        # Should fail when starting node with non-existing database CSV file
        with runner.isolated_filesystem():
            result = runner.invoke(cli_node_start, ["--name", "some-name"])
        self.assertEqual(result.exit_code, 1)

        # now do it with a SQL database which doesn't have to be an existing file
        ctx.databases = [{"label": "some_label", "uri": "data.db", "type": "sql"}]
        with runner.isolated_filesystem():
            result = runner.invoke(cli_node_start, ["--name", "some-name"])
        self.assertEqual(result.exit_code, 0)

    def _setup_stop_test(self, containers):
        container1 = MagicMock()
        container1.name = f"{APPNAME}-iknl-user"
        containers.list.return_value = [container1]

    @patch("docker.DockerClient.containers")
    @patch("vantage6.cli.node.stop.check_docker_running", return_value=True)
    @patch("vantage6.cli.node.stop.NodeContext")
    @patch("vantage6.cli.node.stop.delete_volume_if_exists")
    def test_stop(self, delete_volume, node_context, check_docker, containers):
        self._setup_stop_test(containers)

        runner = CliRunner()

        result = runner.invoke(cli_node_stop, ["--name", "iknl"])

        self.assertEqual(
            result.output, "[info ] - Stopped the vantage6-iknl-user Node.\n"
        )

        self.assertEqual(result.exit_code, 0)

    @patch("docker.DockerClient.containers")
    @patch("vantage6.cli.node.stop.check_docker_running", return_value=True)
    @patch("vantage6.cli.node.stop.NodeContext")
    @patch("vantage6.cli.node.stop.delete_volume_if_exists")
    @patch("vantage6.cli.node.restart.subprocess.run")
    def test_restart(
        self, subprocess_run, delete_volume, node_context, check_docker, containers
    ):
        """Restart a node without errors."""
        self._setup_stop_test(containers)
        # The subprocess.run() function is called with the command to start the node.
        # Unfortunately it is hard to test this, so we just return a successful run
        subprocess_run.return_value = MagicMock(returncode=0)
        runner = CliRunner()
        with runner.isolated_filesystem():
            result = runner.invoke(cli_node_restart, ["--name", "iknl"])
        self.assertEqual(result.exit_code, 0)

    @patch("vantage6.cli.node.attach.attach_logs")
    def test_attach(self, attach_logs):
        """Attach docker logs without errors."""
        runner = CliRunner()
        runner.invoke(cli_node_attach)
        attach_logs.assert_called_once_with("app=node")

    @patch("vantage6.cli.node.clean.q")
    @patch("docker.DockerClient.volumes")
    @patch("vantage6.cli.node.clean.check_docker_running", return_value=True)
    def test_clean(self, check_docker, volumes, q):
        """Clean Docker volumes without errors."""
        volume1 = MagicMock()
        volume1.name = "some-name-tmpvol"
        volumes.list.return_value = [volume1]

        question = MagicMock(name="pop-the-question")
        question.ask.return_value = True
        q.confirm.return_value = question

        runner = CliRunner()
        result = runner.invoke(cli_node_clean)

        # check exit code
        self.assertEqual(result.exit_code, 0)

    @patch("vantage6.cli.node.create_private_key.create_client_and_authenticate")
    @patch("vantage6.cli.node.common.NodeContext")
    @patch("vantage6.cli.node.create_private_key.NodeContext")
    def test_create_private_key(self, context, common_context, client):
        common_context.config_exists.return_value = True
        context.return_value.type_data_folder.return_value = Path(".")
        client.return_value = MagicMock(whoami=MagicMock(organization_name="Test"))
        # client.whoami.organization_name = "Test"
        runner = CliRunner()

        result = runner.invoke(cli_node_create_private_key, ["--name", "application"])

        self.assertEqual(result.exit_code, 0)

        # remove the private key file again
        os.remove("privkey_Test.pem")

    @patch("vantage6.cli.node.create_private_key.RSACryptor")
    @patch("vantage6.cli.node.create_private_key.create_client_and_authenticate")
    @patch("vantage6.cli.node.common.NodeContext")
    @patch("vantage6.cli.node.create_private_key.NodeContext")
    def test_create_private_key_overwite(
        self, context, common_context, client, cryptor
    ):
        common_context.config_exists.return_value = True
        context.return_value.type_data_folder.return_value = Path(".")
        client.return_value = MagicMock(whoami=MagicMock(organization_name="Test"))
        cryptor.create_public_key_bytes.return_value = b""
        # client.whoami.organization_name = "Test"

        runner = CliRunner()

        # overwrite
        with runner.isolated_filesystem():
            with open("privkey_iknl.pem", "w") as f:
                f.write("does-not-matter")

            result = runner.invoke(
                cli_node_create_private_key,
                ["--name", "application", "--overwrite", "--organization-name", "iknl"],
            )
        self.assertEqual(result.exit_code, 0)

        # do not overwrite
        with runner.isolated_filesystem():
            with open("privkey_iknl.pem", "w") as f:
                f.write("does-not-matter")

            result = runner.invoke(
                cli_node_create_private_key,
                ["--name", "application", "--organization-name", "iknl"],
            )

            # print(result.output)

        self.assertEqual(result.exit_code, 0)

    @patch("vantage6.cli.node.common.NodeContext")
    def test_create_private_key_config_not_found(self, context):
        context.config_exists.return_value = False

        runner = CliRunner()
        result = runner.invoke(cli_node_create_private_key, ["--name", "application"])

        self.assertEqual(result.exit_code, 1)

    @patch("vantage6.cli.node.clean.q")
    @patch("docker.DockerClient.volumes")
    @patch("vantage6.common.docker.addons.check_docker_running")
    def test_clean_docker_error(self, check_docker, volumes, q):
        volume1 = MagicMock()
        volume1.name = "some-name-tmpvol"
        volume1.remove.side_effect = APIError("Testing")
        volumes.list.return_value = [volume1]
        question = MagicMock(name="pop-the-question")
        question.ask.return_value = True
        q.confirm.return_value = question

        runner = CliRunner()
        result = runner.invoke(cli_node_clean)

        # check exit code
        self.assertEqual(result.exit_code, 1)

    def test_print_log_worker(self):
        stream = BytesIO("Hello!".encode(STRING_ENCODING))
        temp_stdout = StringIO()
        with contextlib.redirect_stdout(temp_stdout):
            print_log_worker(stream)
        output = temp_stdout.getvalue().strip()
        self.assertEqual(output, "Hello!")

    @patch("vantage6.cli.node.common.info")
    @patch("vantage6.cli.node.common.debug")
    @patch("vantage6.cli.node.common.error")
    @patch("vantage6.cli.node.common.UserClient")
    def test_client(self, client, error, debug, info):
        ctx = MagicMock(
            config={
                "server_url": "localhost",
                "port": Ports.DEV_SERVER.value,
                "api_path": "",
            }
        )

        # should not trigger an exception
        try:
            create_client_and_authenticate(ctx)
        except Exception:
            self.fail("Raised an exception!")

        # client raises exception
        client.side_effect = Exception("Boom!")
        with self.assertRaises(Exception):
            create_client_and_authenticate(ctx)

    # TODO this function has been moved to the common package. A test should
    # be added there instead of here
    # @patch("vantage6.cli.node.error")
    # def test_check_docker(self, error):
    #     docker = MagicMock()
    #     try:
    #         check_docker_running()
    #     except Exception:
    #         self.fail("Exception raised!")

    #     docker.ping.side_effect = Exception("Boom!")
    #     with self.assertRaises(SystemExit):
    #         check_docker_running()



import importlib
import inspect
import json
import os
import sys
from inspect import getmembers, isfunction, ismodule, signature
from pathlib import Path
from types import ModuleType, UnionType
from typing import Any, OrderedDict

import click
import pandas as pd
import questionary as q

from vantage6.common import error, info, warning
from vantage6.common.algorithm_function import (
    get_vantage6_decorator_type,
    is_vantage6_algorithm_func,
)
from vantage6.common.enum import AlgorithmArgumentType, AlgorithmStepType, StrEnumBase

from vantage6.algorithm.client import AlgorithmClient
from vantage6.algorithm.preprocessing.algorithm_json_data import (
    PREPROCESSING_FUNCTIONS_JSON_DATA,
)


class MergePreference:
    """Singleton class to manage global merge preference state"""

    _instance = None
    _prefer_existing = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super(MergePreference, cls).__new__(cls)
        return cls._instance

    @classmethod
    def get_preference(cls) -> bool | None:
        """Get the current merge preference"""
        return cls._prefer_existing

    @classmethod
    def set_preference(cls, prefer_existing: bool) -> None:
        """Set the merge preference globally"""
        cls._prefer_existing = prefer_existing

    @classmethod
    def reset(cls) -> None:
        """Reset the preference to None"""
        cls._prefer_existing = None


class FunctionArgumentType(StrEnumBase):
    """Type of the function argument"""

    PARAMETER = "parameter"
    DATAFRAME = "dataframe"


class Function:
    """Class to handle a function and its JSON representation"""

    def __init__(self, func: callable):
        self.func = func
        self.name = func.__name__
        self.signature = signature(func)
        self.docstring = func.__doc__
        self.json = None
        self.step_type = None

    def prepare_json(self) -> None:
        """Convert the function to a JSON format"""
        self.step_type = self._get_step_type()
        function_json = {
            "name": self.name,
            "display_name": self._pretty_print_name(self.name),
            "standalone": True,
            "description": self._extract_headline_of_docstring(),
            "step_type": self.step_type.value if self.step_type else None,
            "ui_visualizations": [],
            "arguments": [],
            "databases": [],
        }

        parameters = OrderedDict(self.signature.parameters)

        # if the function is a data extraction function, the first argument is a dict
        # with database connection details. This argument should not be added to the
        # function json. Instead, a database should be added to the function json.
        if self.step_type == AlgorithmStepType.DATA_EXTRACTION:
            function_json["databases"].append(
                {
                    "name": "Database",
                    "description": "Database to extract data from",
                }
            )
            # remove database connection details from the signature
            parameters.popitem(last=False)

        # add the arguments to the function json
        for name, param in parameters.items():
            arg_json, arg_type = self._get_argument_json(name, param)
            if arg_json is None:
                continue
            elif arg_type == FunctionArgumentType.DATAFRAME:
                function_json["databases"].append(arg_json)
            else:
                function_json["arguments"].append(arg_json)
        self.json = function_json

    def merge_with_template_json_data(self) -> None:
        """
        Merge the function jsons with the json data from the algorithm_json_data module
        """
        # Only merge the function jsons with template json data if it is an
        # infrastructure-defined function
        if (
            not self.func.__module__.startswith("vantage6.algorithm.")
            or self.json["name"] not in PREPROCESSING_FUNCTIONS_JSON_DATA
        ):
            return

        # get the template json data for the function
        template_json = PREPROCESSING_FUNCTIONS_JSON_DATA[self.json["name"]]
        # merge the dicts, with the template dict taking precedence
        for argument in self.json["arguments"]:
            if argument["name"] in template_json["arguments"]:
                argument.update(template_json["arguments"][argument["name"]])
        # Add any frontend arguments specified in the template json
        if "frontend_arguments" in template_json:
            for frontend_argument in template_json["frontend_arguments"]:
                self._add_frontend_argument(template_json, frontend_argument)

    def merge_with_existing_json(self, existing_json: dict) -> None:
        """Merge the function json with the existing json data"""
        self._merge_dicts(self.json, existing_json)

    def _merge_dicts(self, target: dict, source: dict) -> None:
        """
        Recursively merge source dict into target dict, with source taking precedence
        """
        for key, value in source.items():
            if key in target:
                if isinstance(value, dict) and isinstance(target[key], dict):
                    # Recursively merge nested dictionaries
                    self._merge_dicts(target[key], value)
                else:
                    self._replace_target_with_source(target, key, value)
            else:
                # Add new key-value pair from source to target
                self._replace_target_with_source(target, key, value)

    def _replace_target_with_source(self, target: dict, key: str, value: Any) -> None:
        """Replace the value in target with the one from source"""
        if target[key] == value:
            return

        prefer_existing = MergePreference.get_preference()
        if prefer_existing:
            target[key] = value
        elif prefer_existing is None:
            info(
                f"Different values for the same key '{key}' in function '{self.name}' "
                "were found."
            )
            info(f"Value from function itself: {target[key]}")
            info(f"Value from algorithm.json: {value}")
            result = q.select(
                "Please select the value to keep:",
                choices=[
                    "function itself",
                    "algorithm.json",
                    "function itself (also for all other conflicts)",
                    "algorithm.json (also for all other conflicts)",
                ],
            ).unsafe_ask()
            if result == "algorithm.json":
                target[key] = value
            elif result == "function itself":
                pass  # do nothing
            elif result == "function itself (also for all other conflicts)":
                MergePreference.set_preference(False)
            elif result == "algorithm.json (also for all other conflicts)":
                MergePreference.set_preference(True)
                target[key] = value

    def _get_argument_json(
        self, name: str, param: inspect.Parameter
    ) -> tuple[dict | None, FunctionArgumentType | None]:
        """Get the argument JSON"""

        if param.annotation is None:
            error(f"Function {self.name} has no annotation for argument {name}")
            info(f"Please add a type annotation to the argument {name}")
            info(f"For example, for string arguments: 'def {self.name}({name}: str)'")
            exit(1)

        if param.annotation is AlgorithmClient:
            # Algorithm client arguments do not have to be provided by the user
            return None, None
        elif param.annotation is pd.DataFrame:
            # this is an argument that requires the user to supply a dataframe. That
            # only requires a name and description.
            return {
                "name": name if name != "df" else "Data to use",
                "description": self._extract_parameter_description(name),
            }, FunctionArgumentType.DATAFRAME
        else:
            # This is a regular function parameter
            type_ = self._get_argument_type(param, name)
            arg_json = {
                "name": name,
                "display_name": self._pretty_print_name(name),
                "description": self._extract_parameter_description(name),
                "type": type_.value if type_ else None,
                "required": param.default == inspect.Parameter.empty,
                "has_default_value": param.default != inspect.Parameter.empty,
                "is_frontend_only": False,
            }
            if param.default != inspect.Parameter.empty:
                arg_json["default"] = param.default
            return arg_json, FunctionArgumentType.PARAMETER

    def _add_frontend_argument(
        self, template_json: dict, frontend_argument: str
    ) -> None:
        """Add a frontend argument to the function json"""
        frontend_argument_json: dict = template_json["frontend_arguments"][
            frontend_argument
        ]
        before_arg_name = frontend_argument_json.pop("before_argument")

        try:
            before_arg_idx = next(
                idx
                for idx, arg in enumerate(self.json["arguments"])
                if arg["name"] == before_arg_name
            )
            self.json["arguments"].insert(before_arg_idx, frontend_argument_json)
        except StopIteration:
            warning(
                f"Could not find argument {before_arg_name} in function "
                f"{self.json['name']}. Frontend argument {frontend_argument} "
                "will not be added."
            )

    def _get_argument_type(
        self, param: inspect.Parameter, name: str
    ) -> AlgorithmArgumentType | None:
        """Get the type of the argument"""
        if isinstance(param.annotation, UnionType):
            # Arguments with default values may have type 'str | None'. If that is the
            # case, we want to use the type of the first element in the union.
            if len(param.annotation.__args__) > 2:
                # if there are more than 2 elements in the union, don't handle
                warning(
                    f"Unsupported argument type: {param.annotation} for argument {name}"
                    f" in function {self.name}"
                )
                return None
            elif len(param.annotation.__args__) == 2:
                # if there are two, we want to use the first one if the second is None
                if param.annotation.__args__[1] is type(None):
                    type_ = param.annotation.__args__[0]
                else:
                    warning(
                        f"Unsupported argument type: {param.annotation} for argument "
                        f"{name} in function {self.name}"
                    )
                    return None
            else:
                # normally, unions have 2+ elements. If there is only one, use that
                type_ = param.annotation.__args__[0]
        else:
            type_ = param.annotation

        if type_ == str:
            return AlgorithmArgumentType.STRING
        elif type_ == dict:
            return AlgorithmArgumentType.JSON
        elif type_ == int:
            return AlgorithmArgumentType.INTEGER
        elif type_ == float:
            return AlgorithmArgumentType.FLOAT
        elif type_ == bool:
            return AlgorithmArgumentType.BOOLEAN
        elif type_ == list:
            return AlgorithmArgumentType.STRINGS
        elif type_ == list[str]:
            return AlgorithmArgumentType.STRINGS
        elif type_ == list[int]:
            return AlgorithmArgumentType.INTEGERS
        elif type_ == list[float]:
            return AlgorithmArgumentType.FLOATS
        else:
            warning(
                f"Unsupported argument type: {param.annotation} for argument {name} "
                f"in function {self.name}"
            )
            return None

    def _pretty_print_name(self, name: str) -> str:
        """Pretty print the name of the function"""
        pretty = name.replace("_", " ")
        if len(pretty):
            pretty = pretty[0].upper() + pretty[1:]
        return pretty

    def _extract_headline_of_docstring(self) -> str:
        """Extract the headline of the docstring"""
        if not self.docstring:
            return ""

        # Split by double newlines to get the first paragraph
        paragraphs = self.docstring.split("\n\n")
        first_paragraph = paragraphs[0]

        # Split by single newlines and join the lines with spaces
        lines = first_paragraph.split("\n")
        header = " ".join(line.strip() for line in lines if line.strip() != "")
        return header

    def _get_step_type(self) -> AlgorithmStepType | None:
        """Get the step type of the function"""
        decorator_type = get_vantage6_decorator_type(self.func)
        if decorator_type in AlgorithmStepType.list():
            return decorator_type
        else:
            warning(
                f"Unsupported decorator type: {decorator_type} for function {self.name}"
            )
            return None

    def _extract_parameter_description(self, name: str) -> str:
        """Extract the description of the parameter"""
        if not self.docstring:
            return ""

        # Try both patterns: "{name}:" and "{name} :"
        patterns = [f"{name}:", f"{name} :"]

        for pattern in patterns:
            if pattern in self.docstring:
                return self.docstring.split(pattern)[1].split("\n")[1].strip()

        return ""


@click.command()
@click.option(
    "--algo-function-file",
    default=None,
    type=str,
    help="Path to the file containing or importing the algorithm functions",
)
@click.option(
    "--current-json",
    default=None,
    type=str,
    help="Path to the current algorithm.json file",
)
@click.option(
    "--output-file",
    default="new-algorithm.json",
    type=str,
    help="Path to the output file",
)
def cli_algorithm_generate_algorithm_json(
    algo_function_file: str, current_json: str, output_file: str
) -> dict:
    """
    Generate an updated algorithm.json file to submit to the algorithm store.

    You should provide the path to the file where the algorithm functions are
    defined.

    Note that if you do asterisk ('from x import *') imports, all functions from the
    imported module will be added to the algorithm.json file.
    """
    algo_function_file = _get_algo_function_file_location(algo_function_file)

    current_json = _get_current_json_location(current_json)

    # read the current algorithm.json file
    with open(current_json, "r", encoding="utf-8") as f:
        current_json_data = json.load(f)

    # get the functions from the file
    info(f"Importing functions from {algo_function_file}...")
    functions = _get_functions_from_file(algo_function_file)
    function_objs = [Function(f) for f in functions]

    info("Converting functions to JSON...")
    for function in function_objs:
        function.prepare_json()
        function.merge_with_template_json_data()

        # merge the function jsons with the existing json data
        current_json_func = [
            f for f in current_json_data["functions"] if f["name"] == function.name
        ]
        if current_json_func:
            function.merge_with_existing_json(current_json_func[0])

    # write the new algorithm.json file
    info(f"Writing new algorithm.json file to {output_file}...")
    current_json_data["functions"] = [f.json for f in function_objs]
    with open(output_file, "w", encoding="utf-8") as f:
        json.dump(current_json_data, f, indent=2)

    info(f"New algorithm.json file written to {output_file}")

    warning("-" * 80)
    warning("Always check the generated algorithm.json file before submitting it to ")
    warning("the algorithm store!")
    warning("-" * 80)


def _get_functions_from_file(file_path: str) -> None:
    """Get the functions from the file

    Parameters
    ----------
    file_path : str
        Path to the file containing or importing the algorithm functions
    """
    # Convert path to absolute path
    file_path = str(Path(file_path).resolve())

    # Get the package root directory (two levels up from the file)
    package_root = str(Path(file_path).parent.parent)
    if package_root not in sys.path:
        sys.path.insert(0, package_root)

    # Get the module name from the file path, including the package name
    package_name = Path(file_path).parent.name
    module_name = f"{package_name}.{Path(file_path).stem}"

    # Import the module
    try:
        module = importlib.import_module(module_name)
    except ImportError as e:
        raise ImportError(f"Could not import module {module_name}: {str(e)}") from e

    def get_members_from_module(module: ModuleType) -> list:
        """Get the functions from the module"""
        return [
            member for name, member in getmembers(module) if not name.startswith("_")
        ]

    # get the functions from the algorithm module
    import_members = get_members_from_module(module)
    import_functions = [
        m for m in import_members if isfunction(m) and is_vantage6_algorithm_func(m)
    ]
    import_modules = [m for m in import_members if ismodule(m)]

    # add the functions from the imported modules (only 1 level deep). This is so that
    # if you do e.g. 'from vantage6.algorithm.preprocessing import *', all functions
    # from within those modules are also imported.
    for import_module in import_modules:
        second_level_import_members = get_members_from_module(import_module)
        import_functions.extend(
            [
                m
                for m in second_level_import_members
                if isfunction(m) and is_vantage6_algorithm_func(m)
            ]
        )

    return import_functions


def _get_algo_function_file_location(algo_function_file: str | None) -> None:
    """Get user input for the algorithm creation

    Parameters
    ----------
    algo_function_file : str
        Path to the file containing or importingthe algorithm functions
    """
    if not algo_function_file:
        default_dir = str(Path(os.getcwd()) / "__init__.py")
        algo_function_file = q.text(
            "Path to the file containing or importing the algorithm functions:",
            default=default_dir,
        ).unsafe_ask()

    # Convert to absolute path using pathlib
    algo_function_file = str(Path(algo_function_file).resolve())

    # check if the file exists
    if not Path(algo_function_file).exists():
        raise FileNotFoundError(f"File {algo_function_file} does not exist")

    return algo_function_file


def _get_current_json_location(current_json: str) -> None:
    """Get user input for the current algorithm.json file

    Parameters
    ----------
    current_json : str
        Path to the current algorithm.json file
    """
    if not current_json:
        default_dir = str(Path(os.getcwd()) / "algorithm_store.json")
        current_json = q.text(
            "Path to the current algorithm.json file:",
            default=default_dir,
        ).unsafe_ask()

    # Convert to absolute path using pathlib
    current_json = str(Path(current_json).resolve())

    # check if the file exists
    if not Path(current_json).exists():
        raise FileNotFoundError(f"File {current_json} does not exist")

    return current_json



# import subprocess
# import itertools
# from shutil import rmtree
# from pathlib import Path

# import click
# import docker
# from colorama import Fore, Style

# from vantage6.cli.context.algorithm_store import AlgorithmStoreContext
# from vantage6.common import info, error
# from vantage6.common.globals import APPNAME
# from vantage6.cli.context.server import ServerContext
# from vantage6.cli.context.node import NodeContext
# from vantage6.cli.server.remove import cli_server_remove
# from vantage6.cli.utils import remove_file
# from vantage6.common.globals import InstanceType

# # from vantage6.cli.dev.utils import get_dev_server_context


# @click.command()
# @click.option("-n", "--name", default=None, help="Name of the configuration.")
# @click.option(
#     "-c",
#     "--config",
#     default=None,
#     help="Path to configuration-file; overrides --name",
# )
# @click.pass_context
# def remove_demo_network(
#     click_ctx: click.Context, name: str | None, config: str | None
# ) -> None:
#     """Remove all related demo network files and folders.

#     Select a server configuration to remove that server and the nodes attached
#     to it.
#     """
#     ctx = get_dev_server_context(config, name)

#     # check that the server is not running
#     client = docker.from_env()
#     running_servers = client.containers.list(
#         filters={"label": f"{APPNAME}-type={InstanceType.SERVER}"}
#     )
#     running_server_names = [server.name for server in running_servers]
#     container_name = f"{APPNAME}-{name}-user-{InstanceType.SERVER}"
#     if container_name in running_server_names:
#         error(
#             f"Server {Fore.RED}{name}{Style.RESET_ALL} is still running! First stop "
#             "the network with 'v6 dev stop-demo-network'."
#         )
#         return

#     # remove the server
#     for handler in itertools.chain(ctx.log.handlers, ctx.log.root.handlers):
#         handler.close()
#     click_ctx.invoke(cli_server_remove, ctx=ctx, force=True)

#     # removing the server import config
#     info("Deleting demo import config file")
#     server_configs = ServerContext.instance_folders(
#         InstanceType.SERVER, ctx.name, system_folders=False
#     )
#     import_config_to_del = Path(server_configs["dev"]) / f"{ctx.name}.yaml"
#     remove_file(import_config_to_del, "import_configuration")

#     # also remove the server folder
#     server_configs = ServerContext.instance_folders(
#         InstanceType.SERVER, ctx.name, system_folders=False
#     )
#     server_folder = server_configs["data"]
#     if server_folder.is_dir():
#         rmtree(server_folder)

#     # remove the store folder
#     store_configs = AlgorithmStoreContext.instance_folders(
#         InstanceType.ALGORITHM_STORE, f"{ctx.name}_store", system_folders=False
#     )
#     store_folder = store_configs["data"]
#     if store_folder.is_dir():
#         rmtree(store_folder)

#     # remove the store config file
#     subprocess.run(
#         [
#             "v6",
#             "algorithm-store",
#             "remove",
#             "-n",
#             f"{ctx.name}_store",
#             "--force",
#             "--user",
#         ]
#     )

#     # remove the nodes
#     configs, _ = NodeContext.available_configurations(system_folders=False)
#     node_names = [
#         config.name for config in configs if config.name.startswith(f"{ctx.name}_node_")
#     ]
#     for name in node_names:
#         node_ctx = NodeContext(name, False)
#         for handler in itertools.chain(
#             node_ctx.log.handlers, node_ctx.log.root.handlers
#         ):
#             handler.close()
#         subprocess.run(["v6", "node", "remove", "-n", name, "--user", "--force"])

#     # remove data files attached to the network
#     data_dirs_nodes = NodeContext.instance_folders("node", "", False)["dev"]
#     rmtree(Path(data_dirs_nodes / ctx.name))



import click

from vantage6.common import info
from vantage6.cli.common.utils import attach_logs


@click.command()
def cli_server_attach() -> None:
    """
    Show the server logs in the current console.
    """
    info("Attaching to server logs...")
    attach_logs("app=vantage6-server", "component=vantage6-server")



from __future__ import annotations

from vantage6.common.globals import APPNAME, InstanceType

from vantage6.cli import __version__
from vantage6.cli.configuration_manager import AlgorithmStoreConfigurationManager
from vantage6.cli.context.base_server import BaseServerContext
from vantage6.cli.globals import (
    DEFAULT_SERVER_SYSTEM_FOLDERS as S_FOL,
    ServerType,
)


class AlgorithmStoreContext(BaseServerContext):
    """
    A context class for the algorithm store server.

    Parameters
    ----------
    instance_name : str
        Name of the configuration instance, corresponds to the filename
        of the configuration file.
    system_folders : bool, optional
        System wide or user configuration, by default S_FOL
    """

    INST_CONFIG_MANAGER = AlgorithmStoreConfigurationManager

    def __init__(self, instance_name: str, system_folders: bool = S_FOL):
        super().__init__(
            InstanceType.ALGORITHM_STORE,
            instance_name,
            system_folders=system_folders,
        )
        self.log.info("vantage6 version '%s'", __version__)

    def get_database_uri(self) -> str:
        """
        Obtain the database uri from the environment or the configuration.

        Returns
        -------
        str
            string representation of the database uri
        """
        return super().get_database_uri()

    @property
    def docker_container_name(self) -> str:
        """
        Name of the docker container that the server is running in.

        Returns
        -------
        str
            Server's docker container name
        """
        return f"{APPNAME}-{self.name}-{self.scope}-{ServerType.ALGORITHM_STORE}"

    @classmethod
    def from_external_config_file(
        cls, path: str, system_folders: bool = S_FOL, in_container: bool = False
    ) -> AlgorithmStoreContext:
        """
        Create a server context from an external configuration file. External
        means that the configuration file is not located in the default folders
        but its location is specified by the user.

        Parameters
        ----------
        path : str
            Path of the configuration file
        system_folders : bool, optional
            System wide or user configuration, by default S_FOL
        in_container : bool, optional
            Whether the application is running inside a container, by default False

        Returns
        -------
        AlgorithmStoreContext
            Server context object
        """
        return super().from_external_config_file(
            path,
            ServerType.ALGORITHM_STORE,
            system_folders,
            in_container,
        )

    @classmethod
    def config_exists(cls, instance_name: str, system_folders: bool = S_FOL) -> bool:
        """
        Check if a configuration file exists.

        Parameters
        ----------
        instance_name : str
            Name of the configuration instance, corresponds to the filename
            of the configuration file.
        system_folders : bool, optional
            System wide or user configuration, by default S_FOL

        Returns
        -------
        bool
            Whether the configuration file exists or not
        """
        return super().config_exists(
            InstanceType.ALGORITHM_STORE, instance_name, system_folders
        )

    @classmethod
    def available_configurations(
        cls, system_folders: bool = S_FOL
    ) -> tuple[list, list]:
        """
        Find all available server configurations in the default folders.

        Parameters
        ----------
        system_folders : bool, optional
            System wide or user configuration, by default S_FOL

        Returns
        -------
        tuple[list, list]
            The first list contains validated configuration files, the second
            list contains invalid configuration files.
        """
        return super().available_configurations(
            InstanceType.ALGORITHM_STORE, system_folders
        )



import click

from vantage6.common import info
from vantage6.common.globals import (
    InstanceType,
    Ports,
)

from vantage6.cli.common.decorator import click_insert_context
from vantage6.cli.common.start import (
    helm_install,
    start_port_forward,
)
from vantage6.cli.common.utils import (
    attach_logs,
    create_directory_if_not_exists,
    select_context_and_namespace,
)
from vantage6.cli.context.algorithm_store import AlgorithmStoreContext
from vantage6.cli.globals import ChartName


@click.command()
@click.option("--context", default=None, help="Kubernetes context to use")
@click.option("--namespace", default=None, help="Kubernetes namespace to use")
@click.option("--ip", default=None, help="IP address to listen on")
@click.option("-p", "--port", default=None, type=int, help="Port to listen on")
@click.option(
    "--attach/--detach",
    default=False,
    help="Print server logs to the console after start",
)
@click_insert_context(InstanceType.ALGORITHM_STORE)
def cli_algo_store_start(
    ctx: AlgorithmStoreContext,
    context: str,
    namespace: str,
    ip: str,
    port: int,
    attach: bool,
) -> None:
    """
    Start the algorithm store.
    """
    info("Starting algorithm store...")
    context, namespace = select_context_and_namespace(
        context=context,
        namespace=namespace,
    )

    create_directory_if_not_exists(ctx.log_dir)

    helm_install(
        release_name=ctx.helm_release_name,
        chart_name=ChartName.ALGORITHM_STORE,
        values_file=ctx.config_file,
        context=context,
        namespace=namespace,
    )

    # port forward for server
    info("Port forwarding for algorithm store")
    start_port_forward(
        service_name=f"{ctx.helm_release_name}-vantage6-algorithm-store-service",
        service_port=ctx.config["store"].get("port", Ports.DEV_ALGO_STORE.value),
        port=port or ctx.config["store"].get("port", Ports.DEV_ALGO_STORE.value),
        ip=ip,
        context=context,
        namespace=namespace,
    )

    if attach:
        attach_logs("app=store", "component=store-server")



import os
from os import PathLike
from pathlib import Path

import questionary
import yaml
from colorama import Fore, Style
from kubernetes import config

from vantage6.common import info, warning

from vantage6.cli.globals import (
    DEFAULT_CLI_CONFIG_FILE,
)


class CliConfig:
    """
    A class to manage CLI configuration for Kubernetes context and namespace.

    The CLI configuration is stored in a `config.yaml` file located at the path
    specified by `DEFAULT_CLI_CONFIG_FILE`.

    The `config.yaml` file has the following structure:

    ```yaml
    kube:
      last_context: <last_used_k8s_context>
      last_namespace: <last_used_k8s_namespace>
    ```

    Attributes
    ----------
    config_path : PathLike
        Path to the configuration file.
    _cached_config : dict or None
        Cached configuration data.
    _cached_mtime : float or None
        Last modification time of the configuration file.
    """

    def __init__(self, config_path: str | PathLike = DEFAULT_CLI_CONFIG_FILE) -> None:
        """
        Initialize the CliConfig object.

        Parameters
        ----------
        config_path : str or PathLike, optional
            Path to the configuration file.
        """
        self.config_path: PathLike = Path(config_path)
        self._cached_config: dict | None = None
        self._cached_mtime: float | None = None

    def _load_config(self) -> dict:
        """
        Load the configuration from the file.

        Returns
        -------
        dict
            The loaded configuration data.
        """
        if self.config_path.exists():
            with open(self.config_path, "r") as config_file:
                return yaml.safe_load(config_file) or {}
        return {}

    def _save_config(self, config: dict) -> None:
        """
        Save the configuration to the file.

        Parameters
        ----------
        config : dict
            The configuration data to save.
        """
        self.config_path.parent.mkdir(parents=True, exist_ok=True)
        with open(self.config_path, "w") as config_file:
            yaml.dump(config, config_file)

    def _reload_cache_lazy(self) -> None:
        """
        Reload the configuration cache if the file has been modified.
        """
        if self.config_path.exists():
            mtime = os.path.getmtime(self.config_path)
            if self._cached_mtime != mtime:
                self._cached_config = self._load_config()
                self._cached_mtime = mtime
        else:
            self._cached_config = {}
            self._cached_mtime = None

    def get_last_context(self) -> str | None:
        """
        Get the last used Kubernetes context.

        Returns
        -------
        str or None
            The last used Kubernetes context, or None if not set.
        """
        self._reload_cache_lazy()
        return self._cached_config.get("kube", {}).get("last_context")

    def set_last_context(self, context: str) -> None:
        """
        Set the Kubernetes context.

        Parameters
        ----------
        context : str
            The Kubernetes context to set.
        """
        config = self._load_config()
        if "kube" not in config:
            config["kube"] = {}

        if config["kube"].get("last_context") != context:
            config["kube"]["last_context"] = context
            self._save_config(config)
            self._reload_cache_lazy()

    def get_last_namespace(self) -> str | None:
        """
        Get the last used Kubernetes namespace.

        Returns
        -------
        str or None
            The last used Kubernetes namespace, or None if not set.
        """
        self._reload_cache_lazy()
        return self._cached_config.get("kube", {}).get("last_namespace")

    def set_last_namespace(self, namespace: str) -> None:
        """
        Set the Kubernetes namespace.

        Parameters
        ----------
        namespace : str
            The Kubernetes namespace to set.
        """
        config = self._load_config()
        if "kube" not in config:
            config["kube"] = {}

        if config["kube"].get("last_namespace") != namespace:
            config["kube"]["last_namespace"] = namespace
            self._save_config(config)
            self._reload_cache_lazy()

    def remove_kube(self) -> None:
        """
        Remove the last used context and namespace.
        """
        if self.config_path.exists():
            config = self._load_config()
            if "kube" in config:
                del config["kube"]
                self._save_config(config)
                self._reload_cache_lazy()

    def get_active_settings(
        self,
        context: str | None,
        namespace: str | None,
    ) -> tuple[str, str]:
        """
        Get the active Kubernetes context and namespace.

        Parameters
        ----------
        context : str or None
            The Kubernetes context to use.
        namespace : str or None
            The Kubernetes namespace to use.

        Returns
        -------
        tuple[str, str]
            A tuple containing the active context and namespace.
        """
        if not context:
            _, active_context = config.list_kube_config_contexts()
            context = active_context["name"]

        if not namespace:
            namespace = active_context["context"].get("namespace", "default")

        return context, namespace

    def compare_changes_config(
        self,
        context: str | None = None,
        namespace: str | None = None,
    ) -> tuple[str, str]:
        """
        Compare active settings with last used settings.

        Parameters
        ----------
        context : str or None, optional
            The Kubernetes context to use.
        namespace : str or None, optional
            The Kubernetes namespace to use.

        Returns
        -------
        tuple[str, str]
            A tuple containing the active context and namespace.
        """

        active_context, active_namespace = self.get_active_settings(context, namespace)
        last_context = self.get_last_context()
        last_namespace = self.get_last_namespace()

        # compare context
        if not last_context:
            self.set_last_context(context=active_context)
        elif last_context != active_context:
            warning("Are you using the correct context?")
            warning(f"Current context: {Fore.YELLOW}{active_context}{Style.RESET_ALL}")
            warning(f"Last    context: {Fore.YELLOW}{last_context}{Style.RESET_ALL}")

            active_context = questionary.select(
                "Which context do you want to use?",
                choices=[active_context, last_context],
                default=active_context,
            ).ask()

            if last_context != active_context:
                self.set_last_context(context=active_context)

        # compare namespace
        if not last_namespace:
            self.set_last_namespace(namespace=active_namespace)
        elif last_namespace != active_namespace:
            warning("Are you using the correct namespace?")
            warning(
                f"Current namespace: {Fore.YELLOW}{active_namespace}{Style.RESET_ALL}"
            )
            warning(
                f"Last    namespace: {Fore.YELLOW}{last_namespace}{Style.RESET_ALL}"
            )

            active_namespace = questionary.select(
                "Which namespace do you want to use?",
                choices=[active_namespace, last_namespace],
                default=active_namespace,
            ).ask()

            if last_namespace != active_namespace:
                self.set_last_namespace(namespace=active_namespace)

        info(f"Using    context: {Fore.YELLOW}{active_context}{Style.RESET_ALL}")
        info(f"Using  namespace: {Fore.YELLOW}{active_namespace}{Style.RESET_ALL}")
        return active_context, active_namespace



import json
import subprocess
from pathlib import Path
from subprocess import Popen
from typing import Iterable

import click
import docker
import questionary as q
from colorama import Fore, Style

from vantage6.common import error, warning
from vantage6.common.globals import APPNAME, STRING_ENCODING, InstanceType

from vantage6.cli.config import CliConfig
from vantage6.cli.context import select_context_class
from vantage6.cli.utils import validate_input_cmd_args


def select_context_and_namespace(
    context: str | None = None,
    namespace: str | None = None,
) -> tuple[str, str]:
    """
    Select the context and namespace to use.

    This uses the CLI config to compare the provided context and namespace with the
    last used context and namespace. If the provided context and namespace are not
    the same as the last used context and namespace, the CLI config is updated.

    Parameters
    ----------
    context : str, optional
        The Kubernetes context to use.
    namespace : str, optional
        The Kubernetes namespace to use.

    Returns
    -------
    tuple[str, str]
        The context and namespace to use
    """
    cli_config = CliConfig()

    return cli_config.compare_changes_config(
        context=context,
        namespace=namespace,
    )


def create_directory_if_not_exists(directory: Path) -> None:
    """
    Create a directory.
    """
    try:
        directory.mkdir(parents=True, exist_ok=True)
    except Exception as e:
        error(f"Failed to create directory {directory}: {e}")
        exit(1)


def find_running_service_names(
    instance_type: InstanceType,
    only_system_folders: bool = False,
    only_user_folders: bool = False,
    context: str | None = None,
    namespace: str | None = None,
) -> list[str]:
    """
    List running Vantage6 servers.

    Parameters
    ----------
    instance_type : InstanceType
        The type of instance to find running services for
    only_system_folders : bool, optional
        Whether to look for system-based services or not. By default False.
    only_user_folders : bool, optional
        Whether to look for user-based services or not. By default False.
    context : str, optional
        The Kubernetes context to use.
    namespace : str, optional
        The Kubernetes namespace to use.

    Returns
    -------
    list[str]
        List of release names that are running
    """
    # Input validation
    validate_input_cmd_args(context, "context name", allow_none=True)
    validate_input_cmd_args(namespace, "namespace name", allow_none=True)
    validate_input_cmd_args(instance_type, "instance type", allow_none=False)
    if only_system_folders and only_user_folders:
        error("Cannot use both only_system_folders and only_user_folders")
        exit(1)

    # Create the command
    command = [
        "helm",
        "list",
        "--output",
        "json",  # Get structured output
    ]

    if context:
        command.extend(["--kube-context", context])

    if namespace:
        command.extend(["--namespace", namespace])
    else:
        command.extend(["--all-namespaces"])

    try:
        result = subprocess.run(
            command,
            capture_output=True,
            text=True,
            check=True,
        )
    except subprocess.CalledProcessError as e:
        error(f"Failed to list Helm releases: {e}")
        return []
    except FileNotFoundError:
        error(
            "Helm command not found. Please ensure Helm is installed and available in "
            "the PATH."
        )
        return []

    try:
        releases = json.loads(result.stdout)
    except json.JSONDecodeError:
        error("Failed to parse Helm output as JSON")
        return []

    # filter services for the vantage6 services that are sought. These have
    # the following pattern:
    # f"{APPNAME}-{name}-{scope}-{instance_type.value}"

    # filter for the instance type
    svc_starts_with = f"{APPNAME}-"
    if only_system_folders:
        svc_ends_with = f"system-{instance_type.value}"
    elif only_user_folders:
        svc_ends_with = f"user-{instance_type.value}"
    else:
        svc_ends_with = f"-{instance_type.value}"

    matching_services = []
    for release in releases:
        release_name = release.get("name", "")

        # Check if this is a Vantage6 server release
        is_matching_service = (
            release_name.startswith(svc_starts_with)
            and release_name.endswith(svc_ends_with)
            and instance_type.value in release_name
        )

        if is_matching_service:
            matching_services.append(release_name)

    return matching_services


def select_running_service(
    running_services: list[str],
    instance_type: InstanceType,
) -> str:
    """
    Select a running service from the list of running services.
    """
    try:
        name = q.select(
            f"Select the {instance_type.value} you wish to inspect:",
            choices=running_services,
        ).unsafe_ask()
    except KeyboardInterrupt:
        error("Aborted by user!")
        exit(1)
    return name


def get_server_name(
    name: str,
    system_folders: bool,
    running_server_names: list[str],
    instance_type: InstanceType,
) -> str:
    """
    Get the full name of a running server.

    Parameters
    ----------
    name : str
        Name of the server to get the full name from
    system_folders : bool
        Whether to use system folders or not
    running_server_names : list[str]
        The names of the running servers
    instance_type : InstanceType
        The type of instance to get the full name from
    """

    if not name:
        if not running_server_names:
            error(
                f"No {instance_type.value}s are running! You can only check the version"
                f" for {instance_type.value}s that are running"
            )
            exit(1)
        try:
            name = q.select(
                f"Select the {instance_type.value} you wish to inspect:",
                choices=running_server_names,
            ).unsafe_ask()
        except KeyboardInterrupt:
            error("Aborted by user!")
            exit(1)
    else:
        post_fix = "system" if system_folders else "user"
        name = f"{APPNAME}-{name}-{post_fix}"
    return name


def get_running_servers(
    client: docker.DockerClient, instance_type: InstanceType
) -> list[str]:
    """Get the running servers of a certain type.

    Parameters
    ----------
    client : docker.DockerClient
        The docker client to use
    instance_type : InstanceType
        The type of instance to get the running servers from

    Returns
    -------
    list[str]
        The names of the running servers
    """
    running_servers = client.containers.list(
        filters={"label": f"{APPNAME}-type={instance_type.value}"}
    )
    return [server.name for server in running_servers]


def get_server_configuration_list(instance_type: InstanceType) -> None:
    """
    Print list of available server configurations.

    Parameters
    ----------
    instance_type : InstanceType
        The type of instance to get the configurations for
    """
    ctx_class = select_context_class(instance_type)

    running_server_names = find_running_service_names(instance_type)
    header = "\nName" + (21 * " ") + "Status" + (10 * " ") + "System/User"

    click.echo(header)
    click.echo("-" * len(header))

    running = Fore.GREEN + "Running" + Style.RESET_ALL
    stopped = Fore.RED + "Not running" + Style.RESET_ALL

    # system folders
    configs, failed_imports_system = ctx_class.available_configurations(
        system_folders=True
    )
    for config in configs:
        status = (
            running
            if f"{APPNAME}-{config.name}-system-{instance_type.value}"
            in running_server_names
            else stopped
        )
        click.echo(f"{config.name:25}{status:25} System ")

    # user folders
    configs, failed_imports_user = ctx_class.available_configurations(
        system_folders=False
    )
    for config in configs:
        status = (
            running
            if f"{APPNAME}-{config.name}-user-{instance_type.value}"
            in running_server_names
            else stopped
        )
        click.echo(f"{config.name:25}{status:25} User   ")

    click.echo("-" * 85)
    if len(failed_imports_system) + len(failed_imports_user):
        warning(
            f"{Fore.RED}Failed imports: "
            f"{len(failed_imports_system) + len(failed_imports_user)}{Style.RESET_ALL}"
        )


def print_log_worker(logs_stream: Iterable[bytes]) -> None:
    """
    Print the logs from the logs stream.

    Parameters
    ----------
    logs_stream : Iterable[bytes]
        Output of the container.attach() method
    """
    for log in logs_stream:
        try:
            print(log.decode(STRING_ENCODING), end="")
        except UnicodeDecodeError:
            print(
                "ERROR DECODING LOGS!!! Printing raw bytes. Please check the logs in "
                "the container."
            )
            print(log)


def get_name_from_container_name(container_name: str) -> str:
    """
    Get the node/server/store name from a container name.

    Parameters
    ----------
    container_name : str
        The name of the container

    Returns
    -------
    str
        The name of the node/server/store
    """
    # Container name is structured as: f"{APPNAME}-{name}-{post_fix}"
    # Take into account that name can contain '-'
    return "-".join(container_name.split("-")[1:-1])


def attach_logs(*labels: list[str]) -> None:
    """
    Attach to the logs of the given labels.

    Parameters
    ----------
    labels : list[str]
        The labels to attach to
    """
    command = ["kubectl", "logs", "--follow", "--selector", ",".join(labels)]
    process = Popen(command, stdout=None, stderr=None)
    process.wait()


def get_main_cli_command_name(instance_type: InstanceType) -> str:
    """
    Get the main CLI command name for a given instance type.

    Parameters
    ----------
    instance_type : InstanceType
        The type of instance to get the main CLI command name for
    """
    if instance_type == InstanceType.SERVER:
        return "server"
    elif instance_type == InstanceType.ALGORITHM_STORE:
        return "algorithm-store"
    elif instance_type == InstanceType.NODE:
        return "node"
    else:
        raise ValueError(f"Invalid instance type: {instance_type}")



# import click
# import json
# import subprocess
# import sys

# from pathlib import Path

# from rich.console import Console

# from vantage6.cli.dev.create import create_demo_network
# from vantage6.cli.dev.remove import remove_demo_network

# # from vantage6.cli.dev.start import start_demo_network
# # from vantage6.cli.dev.stop import stop_demo_network
# from vantage6.cli.utils import prompt_config_name
# from vantage6.common.globals import Ports

# TEST_FILE_PATH = Path(__file__).parent / "algo_test_scripts" / "algo_test_script.py"


# @click.command()
# @click.option(
#     "--script",
#     type=click.Path(),
#     default=TEST_FILE_PATH,
#     help="Path of the script to test the algorithm. If a script is not provided, the default script is used.",
# )
# @click.option(
#     "--task-arguments",
#     type=str,
#     default=None,
#     help="Arguments to be provided to Task.create function. If --script is provided, this should not be set.",
# )
# @click.option(
#     "--create-dev-network",
#     is_flag=True,
#     help="Create a new dev network to run the test",
# )
# @click.option(
#     "--start-dev-network",
#     is_flag=True,
#     help="Start a dev network to run the test",
# )
# @click.option(
#     "-n", "--name", default=None, type=str, help="Name for your development setup"
# )
# @click.option(
#     "--server-url",
#     type=str,
#     default="http://host.docker.internal",
#     help="Server URL to point to. If you are using Docker Desktop, "
#     "the default http://host.docker.internal should not be changed.",
# )
# @click.option(
#     "-i", "--image", type=str, default=None, help="Server Docker image to use"
# )
# @click.option(
#     "--keep",
#     type=bool,
#     default=False,
#     help="Keep the dev network after finishing the test",
# )
# @click.option(
#     "--add-dataset",
#     type=(str, click.Path()),
#     default=[],
#     multiple=True,
#     help="Add a dataset to the nodes. The first argument is the label of the database, "
#     "the second is the path to the dataset file.",
# )
# @click.pass_context
# def cli_test_client_script(
#     click_ctx: click.Context,
#     script: Path | None,
#     task_arguments: str | None,
#     name: str,
#     server_url: str,
#     create_dev_network: bool,
#     start_dev_network: bool,
#     image: str,
#     keep: bool,
#     add_dataset: list[tuple[str, Path]] = (),
# ) -> int:
#     """
#     Run a script for testing an algorithm on a dev network.
#     The path to the script must be provided as an argument.
#     """
#     if not (script or task_arguments):
#         raise click.UsageError("--script or --task-arguments must be set.")
#     elif script != TEST_FILE_PATH and task_arguments:
#         raise click.UsageError("--script and --task-arguments cannot be set together.")

#     # Check if the task_arguments is a valid JSON string
#     if task_arguments:
#         try:
#             json.loads(task_arguments.replace("'", '"'))
#         except json.JSONDecodeError:
#             raise click.UsageError("task-arguments must be a valid JSON string.")

#     name = prompt_config_name(name)

#     # create the network
#     if create_dev_network:
#         click_ctx.invoke(
#             create_demo_network,
#             name=name,
#             num_nodes=3,
#             server_url=server_url,
#             server_port=Ports.DEV_SERVER.value,
#             image=image,
#             extra_server_config=None,
#             extra_node_config=None,
#             add_dataset=add_dataset,
#         )

#     # start the server and nodes
#     if create_dev_network or start_dev_network:
#         click_ctx.invoke(
#             start_demo_network,
#             name=name,
#             server_image=image,
#             node_image=image,
#         )

#     # run the test script and get the result
#     if not task_arguments:
#         subprocess_args = ["python", script]
#     else:
#         subprocess_args = ["python", script, task_arguments]

#     result = subprocess.run(subprocess_args, stdout=sys.stdout, stderr=sys.stderr)

#     # check the exit code. If the test passed, it should be 0
#     if result.returncode == 0:
#         msg = ":heavy_check_mark: [green]Test passed[/green]"
#     else:
#         msg = ":x: [red]Test failed[/red]"

#     console = Console()
#     console.print(msg)

#     # clean up the test resources. Keep the network if --keep is set, or if the network
#     # was created for this test. If the network was started for this test, stop it but
#     # do not remove it.
#     if not keep:
#         if create_dev_network or start_dev_network:
#             click_ctx.invoke(stop_demo_network, name=name)
#         if create_dev_network:
#             click_ctx.invoke(remove_demo_network, name=name)

#     return result.returncode



import click

from vantage6.common.docker.addons import check_docker_running
from vantage6.common.globals import InstanceType

from vantage6.cli.common.utils import get_server_configuration_list


@click.command()
def cli_algo_store_configuration_list() -> None:
    """
    Print the available server configurations.
    """
    check_docker_running()

    get_server_configuration_list(InstanceType.ALGORITHM_STORE)



import click

from vantage6.common import info
from vantage6.cli.common.utils import attach_logs


@click.command()
def cli_node_attach() -> None:
    """
    Show the node logs in the current console.
    """
    info("Attaching to node logs...")
    attach_logs("app=node")



average = {
    "collaboration": 1,
    "organizations": [1],
    "name": "test_average_task",
    "image": "harbor2.vantage6.ai/demo/average",
    "description": "",
    "method": "central_average",
    "arguments": {
        "column_name": "Age",
    },
    "databases": [{"label": "olympic_athletes"}],
}

kaplan_meier = {
    "collaboration": 1,
    "organizations": [1],
    "name": "test_average_task",
    "image": "harbor2.vantage6.ai/algorithms/kaplan-meier",
    "description": "",
    "method": "kaplan_meier_central",
    "arguments": {
        "time_column_name": "days",
        "censor_column_name": "censor",
        "organizations_to_include": [1, 2, 3],
    },
    "databases": [{"label": "kaplan_meier_test"}],
}

args = {"average": average, "kaplan_meier": kaplan_meier}



from colorama import Fore, Style

from vantage6.common import ensure_config_dir_writable, error, info
from vantage6.common.globals import InstanceType

from vantage6.cli.common.utils import get_main_cli_command_name
from vantage6.cli.config import CliConfig
from vantage6.cli.configuration_wizard import configuration_wizard
from vantage6.cli.context import select_context_class
from vantage6.cli.utils import check_config_name_allowed, prompt_config_name


def new(
    name: str, system_folders: bool, namespace: str, context: str, type_: InstanceType
):
    cli_config = CliConfig()
    context, namespace = cli_config.compare_changes_config(
        context=context,
        namespace=namespace,
    )

    name = prompt_config_name(name)

    # check if name is valid
    check_config_name_allowed(name)

    # check that this config does not exist
    ctx_class = select_context_class(type_)
    try:
        if ctx_class.config_exists(name, system_folders):
            error(f"Configuration {Fore.RED}{name}{Style.RESET_ALL} already exists!")
            exit(1)
    except Exception as e:
        error(e)
        exit(1)

    command_name = get_main_cli_command_name(type_)

    # Check that we can write in this folder
    if not ensure_config_dir_writable(system_folders):
        error("Your user does not have write access to all folders. Exiting")
        info(
            f"Create a new {command_name} using '{Fore.GREEN}v6 {command_name} new "
            f"--user{Style.RESET_ALL}' instead!"
        )
        exit(1)

    # create config in ctx location
    try:
        cfg_file = configuration_wizard(type_, name, system_folders)
    except KeyboardInterrupt:
        error("Configuration creation aborted.")
        exit(1)
    info(f"New configuration created: {Fore.GREEN}{cfg_file}{Style.RESET_ALL}")

    flag = "" if system_folders else "--user"
    info(
        f"You can start the {command_name} by running {Fore.GREEN}v6 {command_name} "
        f"start {flag}{Style.RESET_ALL}"
    )



# from pathlib import Path
# import click

# from vantage6.common.globals import Ports
# from vantage6.cli.utils import info
# from vantage6.cli.dev.create import create_demo_network
# from vantage6.cli.dev.start import start_demo_network
# from vantage6.cli.dev.stop import stop_demo_network
# from vantage6.cli.dev.remove import remove_demo_network
# from vantage6.cli.utils import prompt_config_name, check_config_name_allowed
# from vantage6.cli.test.feature_tester import cli_test_features


# @click.command()
# @click.option(
#     "-n", "--name", default=None, type=str, help="Name for your development setup"
# )
# @click.option(
#     "--server-url",
#     type=str,
#     default="http://host.docker.internal",
#     help="Server URL to point to. If you are using Docker Desktop, "
#     "the default http://host.docker.internal should not be changed.",
# )
# @click.option(
#     "-i", "--image", type=str, default=None, help="Server Docker image to use"
# )
# @click.option(
#     "--keep",
#     type=bool,
#     default=False,
#     help="Keep the dev network after finishing the test",
# )
# @click.option(
#     "--extra-server-config",
#     type=click.Path(exists=True),
#     default=None,
#     help="YAML File with additional server "
#     "configuration. This will be appended to the server "
#     "configuration file",
# )
# @click.option(
#     "--extra-node-config",
#     type=click.Path("rb"),
#     default=None,
#     help="YAML File with additional node configuration. This will be"
#     " appended to each of the node configuration files",
# )
# @click.pass_context
# def cli_test_integration(
#     click_ctx: click.Context,
#     name: str,
#     server_url: str,
#     image: str,
#     keep: bool = False,
#     extra_server_config: Path = None,
#     extra_node_config: Path = None,
# ) -> list[dict]:
#     """
#     Create dev network and run diagnostic checks on it.

#     This is a full integration test of the vantage6 network. It will create
#     a test server with some nodes using the `vdev` commands, and then run the
#     v6-diagnostics algorithm to test all functionality.
#     """
#     # get name for the development setup - if not given - and check if it is
#     # allowed
#     name = prompt_config_name(name)
#     check_config_name_allowed(name)

#     # create server & node configurations and create test resources (
#     # collaborations, organizations, etc)
#     click_ctx.invoke(
#         create_demo_network,
#         name=name,
#         num_nodes=3,
#         server_url=server_url,
#         server_port=Ports.DEV_SERVER.value,
#         image=image,
#         extra_server_config=extra_server_config,
#         extra_node_config=extra_node_config,
#     )

#     # start the server and nodes
#     click_ctx.invoke(
#         start_demo_network,
#         name=name,
#         server_image=image,
#         node_image=image,
#     )

#     # run the diagnostic tests
#     # TODO the username and password should be coordinated with the vdev
#     # command - at present it spits out this username/password combination by
#     # default but both should be defined in the same place
#     # TODO VPN testing is always excluded - allow to include it with a flag
#     # when vdev commands can handle extra config parameters
#     diagnose_results = click_ctx.invoke(
#         cli_test_features,
#         host="http://localhost",
#         port=Ports.DEV_SERVER.value,
#         api_path="/api",
#         username="dev_admin",
#         password="password",
#         collaboration=1,
#         organizations=None,
#         all_nodes=True,
#         online_only=False,
#         no_vpn=True,
#     )

#     # clean up the test resources
#     click_ctx.invoke(stop_demo_network, name=name)
#     if not keep:
#         click_ctx.invoke(remove_demo_network, name=name)
#     else:
#         info(
#             f"Keeping the demo network {name}. You can start it with `v6 dev "
#             "start-demo-network`"
#         )

#     return diagnose_results



"""
Utility functions for the CLI
"""

from __future__ import annotations

import os
import re
import subprocess
from pathlib import Path

import docker
import questionary as q

from vantage6.common import error, info, warning


def check_config_name_allowed(name: str) -> None:
    """
    Check if configuration name is allowed

    Parameters
    ----------
    name : str
        Name to be checked
    """
    if name.count(" ") > 0:
        name = name.replace(" ", "-")
        info(f"Replaced spaces from configuration name: {name}")
    if not re.match("^[a-zA-Z0-9_.-]+$", name):
        error(
            f"Name '{name}' is not allowed. Please use only the following "
            "characters: a-zA-Z0-9_.-"
        )
        # FIXME: FM, 2023-01-03: I dont think this is a good side effect. This
        # should be handled by the caller.
        exit(1)


def check_if_docker_daemon_is_running(docker_client: docker.DockerClient) -> None:
    """
    Check if Docker daemon is running

    Parameters
    ----------
    docker_client : docker.DockerClient
        The docker client
    """
    try:
        docker_client.ping()
    except Exception:
        error("Docker socket can not be found. Make sure Docker is running.")
        exit(1)


def remove_file(file: str | Path, file_type: str) -> None:
    """
    Remove a file if it exists.

    Parameters
    ----------
    file : str
        absolute path to the file to be deleted
    file_type : str
        type of file, used for logging
    """
    if os.path.isfile(file):
        info(f"Removing {file_type} file: {file}")
        try:
            os.remove(file)
        except Exception as e:
            error(f"Could not delete file: {file}")
            error(e)
    else:
        warning(f"Could not remove {file_type} file: {file} does not exist")


def prompt_config_name(name: str | None) -> None:
    """
    Get a new configuration name from the user, or simply return the name if
    it is not None.

    Parameters
    ----------
    name : str
        Name to be checked

    Returns
    -------
    str
        The name of the configuration
    """
    if not name:
        try:
            name = q.text("Please enter a configuration-name:").unsafe_ask()
        except KeyboardInterrupt:
            error("Aborted by user!")
            exit(1)
        if name.count(" ") > 0:
            name = name.replace(" ", "-")
            info(f"Replaced spaces from configuration name: {name}")
    return name


def switch_context_and_namespace(
    context: str | None = None, namespace: str | None = None
) -> None:
    # input validation
    validate_input_cmd_args(context, "context name", allow_none=True)
    validate_input_cmd_args(namespace, "namespace name", allow_none=True)

    try:
        if context:
            subprocess.run(
                ["kubectl", "config", "use-context", context],
                check=True,
                stdout=subprocess.DEVNULL,
            )
            info(f"Successfully set context to: {context}")

        if namespace:
            subprocess.run(
                [
                    "kubectl",
                    "config",
                    "set-context",
                    context or "--current",
                    f"--namespace={namespace}",
                ],
                check=True,
                stdout=subprocess.DEVNULL,
            )
            info(f"Successfully set namespace to: {namespace}")

    except subprocess.CalledProcessError as e:
        error(f"Failed to set Kubernetes context or namespace: {e}")


def validate_input_cmd_args(
    value: str | None, field_name: str, allow_none: bool = False
) -> None:
    """
    Validate input for subprocess commands.

    Parameters
    ----------
    value : str | None
        The value to validate.
    field_name : str
        The name of the field being validated, used for error messages.
    allow_none : bool, optional
        Whether None is allowed as a valid value. Defaults to False.

    Raises
    ------
    SystemExit
        If the input is invalid.
    """
    if allow_none and value is None:
        return

    if not isinstance(value, str) or not re.match("^[a-zA-Z0-9_.-]+$", value):
        error(
            f"Invalid {field_name}: {value}. Use only alphanumeric characters, "
            "dashes, underscores, or dots."
        )
        exit(1)



import time

import click
import docker
import questionary as q
from colorama import Fore, Style

from vantage6.common import error, info, warning
from vantage6.common.docker.addons import (
    check_docker_running,
    delete_volume_if_exists,
    get_server_config_name,
    stop_container,
)
from vantage6.common.globals import APPNAME

from vantage6.cli.context import NodeContext
from vantage6.cli.globals import DEFAULT_NODE_SYSTEM_FOLDERS as N_FOL
from vantage6.cli.node.common import find_running_node_names


@click.command()
@click.option("-n", "--name", default=None, help="Configuration name")
@click.option(
    "--system",
    "system_folders",
    flag_value=True,
    help="Search for configuration in system folders instead of user folders",
)
@click.option(
    "--user",
    "system_folders",
    flag_value=False,
    default=N_FOL,
    help="Search for configuration in the user folders instead of "
    "system folders. This is the default.",
)
@click.option("--all", "all_nodes", flag_value=True, help="Stop all running nodes")
@click.option(
    "--force",
    "force",
    flag_value=True,
    help="Kill nodes instantly; don't wait for them to shut down",
)
def cli_node_stop(
    name: str, system_folders: bool, all_nodes: bool, force: bool
) -> None:
    """
    Stop one or all running nodes.
    """
    check_docker_running()
    client = docker.from_env()

    running_node_names = find_running_node_names(client)

    if not running_node_names:
        warning("No nodes are currently running.")
        return

    if force:
        warning(
            "Forcing the node to stop will not terminate helper "
            "containers, neither will it remove routing rules made on the "
            "host!"
        )

    if all_nodes:
        for container_name in running_node_names:
            _stop_node(client, container_name, force, system_folders)
    else:
        if not name:
            try:
                container_name = q.select(
                    "Select the node you wish to stop:", choices=running_node_names
                ).unsafe_ask()
            except KeyboardInterrupt:
                error("Aborted by user!")
                return
        else:
            post_fix = "system" if system_folders else "user"
            container_name = f"{APPNAME}-{name}-{post_fix}"

        if container_name in running_node_names:
            _stop_node(client, container_name, force, system_folders)
            info(f"Stopped the {Fore.GREEN}{container_name}{Style.RESET_ALL} Node.")
        else:
            error(f"{Fore.RED}{name}{Style.RESET_ALL} is not running?!")


def _stop_node(
    client: docker.DockerClient, container_name: str, force: bool, system_folders: bool
) -> None:
    """
    Stop a node

    Parameters
    ----------
    client : docker.DockerClient
        Docker client
    name : str
        Name of the node container to stop
    force : bool
        Whether to force the node to stop
    system_folders : bool
        Whether to use system folders or not
    """
    container = client.containers.get(container_name)
    # Stop the container. Using stop() gives the container 10s to exit
    # itself, if not then it will be killed
    stop_container(container, force)

    # Sleep for 1 second. Not doing so often causes errors that docker volumes deleted
    # below are 'still in use' when you try to remove them a few ms after the container
    # has been removed
    time.sleep(1)

    # Delete volumes. This is done here rather than within the node container when
    # it is stopped, because at that point the volumes are still in use. Here, the node
    # has already been stopped
    scope = "system" if system_folders else "user"
    config_name = get_server_config_name(container_name, scope)
    ctx = NodeContext(config_name, system_folders, print_log_header=False)

    # Do not delete the data volume as this would remove all sessions.
    for volume in [
        # ctx.docker_volume_name,
        ctx.docker_squid_volume_name,
        ctx.docker_ssh_volume_name,
        ctx.docker_vpn_volume_name,
    ]:
        delete_volume_if_exists(client, volume)



from __future__ import annotations

import hashlib
import os.path
from pathlib import Path

from vantage6.common.context import AppContext
from vantage6.common.globals import APPNAME, STRING_ENCODING, InstanceType

from vantage6.cli import __version__
from vantage6.cli.configuration_manager import NodeConfigurationManager
from vantage6.cli.globals import DEFAULT_NODE_SYSTEM_FOLDERS as N_FOL


class NodeContext(AppContext):
    """
    Node context object for the host system.

    Parameters
    ----------
    instance_name : str
        Name of the configuration instance, corresponds to the filename
        of the configuration file.
    system_folders : bool, optional
        _description_, by default N_FOL
    config_file : str, optional
        _description_, by default None
    in_container : bool, optional
        Whether the application is running inside a container, by default False
    """

    # The server configuration manager is aware of the structure of the server
    # configuration file and makes sure only valid configuration can be loaded.
    INST_CONFIG_MANAGER = NodeConfigurationManager

    def __init__(
        self,
        instance_name: str,
        system_folders: bool = N_FOL,
        config_file: str = None,
        print_log_header: bool = True,
        logger_prefix: str = "",
        in_container: bool = False,
    ):
        super().__init__(
            InstanceType.NODE,
            instance_name,
            system_folders=system_folders,
            config_file=config_file,
            print_log_header=print_log_header,
            logger_prefix=logger_prefix,
            in_container=in_container,
        )
        if print_log_header:
            self.log.info("vantage6 version '%s'", __version__)
        self.identifier = self.__create_node_identifier()

    @classmethod
    def from_external_config_file(
        cls, path: str, system_folders: bool = N_FOL
    ) -> NodeContext:
        """
        Create a node context from an external configuration file. External
        means that the configuration file is not located in the default folders
        but its location is specified by the user.

        Parameters
        ----------
        path : str
            Path of the configuration file
        system_folders : bool, optional
            System wide or user configuration, by default N_FOL

        Returns
        -------
        NodeContext
            Node context object
        """
        return super().from_external_config_file(
            Path(path).resolve(), InstanceType.NODE, system_folders
        )

    @classmethod
    def config_exists(cls, instance_name: str, system_folders: bool = N_FOL) -> bool:
        """
        Check if a configuration file exists.

        Parameters
        ----------
        instance_name : str
            Name of the configuration instance, corresponds to the filename
            of the configuration file.
        system_folders : bool, optional
            System wide or user configuration, by default N_FOL

        Returns
        -------
        bool
            Whether the configuration file exists or not
        """
        return super().config_exists(
            InstanceType.NODE, instance_name, system_folders=system_folders
        )

    @classmethod
    def available_configurations(
        cls, system_folders: bool = N_FOL
    ) -> tuple[list, list]:
        """
        Find all available server configurations in the default folders.

        Parameters
        ----------
        system_folders : bool, optional
            System wide or user configuration, by default N_FOL

        Returns
        -------
        tuple[list, list]
            The first list contains validated configuration files, the second
            list contains invalid configuration files.
        """
        return super().available_configurations(InstanceType.NODE, system_folders)

    @staticmethod
    def type_data_folder(system_folders: bool = N_FOL) -> Path:
        """
        Obtain OS specific data folder where to store node specific data.

        Parameters
        ----------
        system_folders : bool, optional
            System wide or user configuration

        Returns
        -------
        Path
            Path to the data folder
        """
        return AppContext.type_data_folder(InstanceType.NODE, system_folders)

    @property
    def databases(self) -> dict:
        """
        Dictionary of local databases that are available for this node.

        Returns
        -------
        dict
            dictionary with database names as keys and their corresponding
            paths as values.
        """
        return self.config["databases"]

    @property
    def docker_container_name(self) -> str:
        """
        Docker container name of the node.

        Returns
        -------
        str
            Node's Docker container name
        """
        return f"{APPNAME}-{self.name}-{self.scope}"

    @property
    def docker_network_name(self) -> str:
        """
        Private Docker network name which is unique for this node.

        Returns
        -------
        str
            Docker network name
        """
        return f"{APPNAME}-{self.name}-{self.scope}-net"

    @property
    def docker_volume_name(self) -> str:
        """
        Docker volume in which task data is stored. In case a file based
        database is used, this volume contains the database file as well.

        Returns
        -------
        str
            Docker volume name
        """
        return os.environ.get("DATA_VOLUME_NAME", f"{self.docker_container_name}-vol")

    @property
    def proxy_log_file(self):
        return self.log_file_name(type_="proxy_server")

    def get_database_uri(self, label: str = "default") -> str:
        """
        Obtain the database URI for a specific database.

        Parameters
        ----------
        label : str, optional
            Database label, by default "default"

        Returns
        -------
        str
            URI to the database
        """
        return self.config["databases"][label]

    def set_folders(
        self, instance_type: InstanceType, instance_name: str, system_folders: bool
    ) -> None:
        """
        Set the folders for the node.

        Parameters
        ----------
        instance_type : InstanceType
            Instance type
        instance_name : str
            Instance name
        system_folders : bool
            Whether to use system folders
        """
        dirs = self.instance_folders(instance_type, instance_name, system_folders)

        self.log_dir = dirs.get("log")
        self.data_dir = dirs.get("data")
        self.config_dir = dirs.get("config")

    @staticmethod
    def instance_folders(*_args, **_kwargs) -> dict:
        """Log, data and config folders are always mounted. The node manager
        should take care of this.
        """
        mnt = Path("/mnt")

        return {
            "log": mnt / "log",
            "data": mnt / "data",
            "config": mnt / "config",
        }

    def __create_node_identifier(self) -> str:
        """
        Create a unique identifier for the node.

        Returns
        -------
        str
            Unique identifier for the node
        """
        return hashlib.sha256(
            os.environ.get("V6_API_KEY").encode(STRING_ENCODING)
        ).hexdigest()[:16]



"""
Common functions that are used in node CLI commands
"""

import os

import docker
from colorama import Fore, Style

from vantage6.common import debug, error, info
from vantage6.common.globals import APPNAME, InstanceType, RequiredNodeEnvVars

from vantage6.client import UserClient

from vantage6.cli.configuration_wizard import select_configuration_questionaire
from vantage6.cli.context.node import NodeContext


def create_client(ctx: NodeContext) -> UserClient:
    """
    Create a client instance.

    Parameters
    ----------
    ctx : NodeContext
        Context of the node loaded from the configuration file
    Returns
    -------
    UserClient
        vantage6 client
    """
    host = ctx.config["server_url"]
    # if the server is run locally, we need to use localhost here instead of
    # the host address of docker
    if host in ["http://host.docker.internal", "http://172.17.0.1"]:
        host = "http://localhost"
    port = ctx.config["port"]
    api_path = ctx.config["api_path"]
    info(f"Connecting to server at '{host}:{port}{api_path}'")
    return UserClient(
        server_url=f"{host}:{port}{api_path}",
        auth_url=os.environ.get(RequiredNodeEnvVars.KEYCLOAK_URL.value),
        log_level="warn",
    )


def create_client_and_authenticate(ctx: NodeContext) -> UserClient:
    """
    Generate a client and authenticate with the server.

    Parameters
    ----------
    ctx : NodeContext
        Context of the node loaded from the configuration file

    Returns
    -------
    UserClient
        vantage6 client
    """
    client = create_client(ctx)

    try:
        client.authenticate()
    except Exception as exc:
        error("Could not authenticate with server!")
        debug(exc)
        exit(1)

    return client


def select_node(name: str, system_folders: bool) -> tuple[str, str]:
    """
    Let user select node through questionnaire if name is not given.

    Returns
    -------
    str
        Name of the configuration file
    """
    name = (
        name
        if name
        else select_configuration_questionaire(InstanceType.NODE, system_folders)
    )

    # raise error if config could not be found
    if not NodeContext.config_exists(name, system_folders):
        error(
            f"The configuration {Fore.RED}{name}{Style.RESET_ALL} could not be found."
        )
        exit(1)
    return name


def find_running_node_names(client: docker.DockerClient) -> list[str]:
    """
    Returns a list of names of running nodes.

    Parameters
    ----------
    client : docker.DockerClient
        Docker client instance

    Returns
    -------
    list[str]
        List of names of running nodes
    """
    running_nodes = client.containers.list(
        filters={"label": f"{APPNAME}-type={InstanceType.NODE.value}"}
    )
    return [node.name for node in running_nodes]



import os.path
import time
from pathlib import Path
from threading import Thread

import click
import docker
from colorama import Fore, Style

from vantage6.common import debug, error, info, warning
from vantage6.common.dataclass import TaskDB
from vantage6.common.docker.addons import (
    check_docker_running,
    remove_container_if_exists,
)
from vantage6.common.globals import (
    APPNAME,
    DEFAULT_DOCKER_REGISTRY,
    DEFAULT_NODE_IMAGE,
    DEFAULT_NODE_IMAGE_WO_TAG,
    InstanceType,
)

from vantage6.cli import __version__
from vantage6.cli.common.decorator import click_insert_context
from vantage6.cli.common.start import pull_infra_image
from vantage6.cli.common.utils import print_log_worker
from vantage6.cli.context.node import NodeContext
from vantage6.cli.node.common import create_client
from vantage6.cli.utils import check_config_name_allowed


@click.command()
@click.option("-i", "--image", default=None, help="Node Docker image to use")
@click.option(
    "--keep/--auto-remove",
    default=False,
    help="Keep node container after finishing. Useful for debugging",
)
@click.option(
    "--force-db-mount",
    is_flag=True,
    help="Always mount node databases; skip the check if they are existing files.",
)
@click.option(
    "--attach/--detach",
    default=False,
    help="Show node logs on the current console after starting the node",
)
@click.option(
    "--mount-src",
    default="",
    help="Override vantage6 source code in container with the source code in this path",
)
@click_insert_context(InstanceType.NODE, include_name=True, include_system_folders=True)
def cli_node_start(
    ctx: NodeContext,
    name: str,
    system_folders: bool,
    image: str,
    keep: bool,
    mount_src: str,
    attach: bool,
    force_db_mount: bool,
) -> None:
    """
    Start the node.
    """
    check_docker_running()
    info("Starting node...")
    info("Finding Docker daemon")
    docker_client = docker.from_env()
    NodeContext.LOGGING_ENABLED = False

    # check if config name is allowed docker name, else exit
    check_config_name_allowed(ctx.name)

    # check that this node is not already running
    running_nodes = docker_client.containers.list(
        filters={"label": f"{APPNAME}-type={InstanceType.NODE.value}"}
    )

    suffix = "system" if system_folders else "user"
    for node in running_nodes:
        if node.name == f"{APPNAME}-{name}-{suffix}":
            error(f"Node {Fore.RED}{name}{Style.RESET_ALL} is already running")
            exit(1)

    # make sure the (host)-task and -log dir exists
    info("Checking that data and log dirs exist")
    ctx.data_dir.mkdir(parents=True, exist_ok=True)
    ctx.log_dir.mkdir(parents=True, exist_ok=True)

    # Determine image-name. First we check if the option --image has been used.
    # Then we check if the image has been specified in the config file, and
    # finally we use the default settings from the package.
    if not image:
        custom_images: dict = ctx.config.get("images")
        if custom_images:
            image = custom_images.get("node")
        else:
            # if no custom image is specified, find the server version and use
            # the latest images from that minor version
            client = create_client(ctx)
            major_minor = None
            try:
                # try to get server version, skip if can't get a connection
                version = client.util.get_server_version(attempts_on_timeout=3)[
                    "version"
                ]
                major_minor = ".".join(version.split(".")[:2])
                image = (
                    f"{DEFAULT_DOCKER_REGISTRY}/"
                    f"{DEFAULT_NODE_IMAGE_WO_TAG}"
                    f":{major_minor}"
                )
            except Exception:
                warning("Could not determine server version. Using default node image")

            if major_minor and not __version__.startswith(major_minor):
                warning(
                    "Version mismatch between CLI and server/node. CLI is "
                    f"running on version {__version__}, while node and server "
                    f"are on version {major_minor}. This might cause "
                    f"unexpected issues; changing to {major_minor}.<latest> "
                    "is recommended."
                )

        # fail safe, in case no custom image is specified and we can't get the
        # server version
        if not image:
            image = f"{DEFAULT_DOCKER_REGISTRY}/{DEFAULT_NODE_IMAGE}"

    info(f"Pulling latest node image '{image}'")
    pull_infra_image(docker_client, image, InstanceType.NODE)

    data_volume = docker_client.volumes.create(ctx.docker_volume_name)
    vpn_volume = docker_client.volumes.create(ctx.docker_vpn_volume_name)
    ssh_volume = docker_client.volumes.create(ctx.docker_ssh_volume_name)
    squid_volume = docker_client.volumes.create(ctx.docker_squid_volume_name)

    info("Creating file & folder mounts")
    # FIXME: should obtain mount points from DockerNodeContext
    mounts = [
        # (target, source)
        ("/mnt/log", str(ctx.log_dir)),
        ("/mnt/data", data_volume.name),
        ("/mnt/vpn", vpn_volume.name),
        ("/mnt/ssh", ssh_volume.name),
        ("/mnt/squid", squid_volume.name),
        ("/mnt/config", str(ctx.config_dir)),
        ("/var/run/docker.sock", "/var/run/docker.sock"),
    ]

    if mount_src:
        # If mount_src is a relative path, docker will consider it a volume.
        mount_src = os.path.abspath(mount_src)
        mounts.append(("/vantage6", mount_src))

    # FIXME: Code duplication: Node.__init__() (vantage6/node/__init__.py)
    #   uses a lot of the same logic. Suggest moving this to
    #   ctx.get_private_key()
    filename = ctx.config.get("encryption", {}).get("private_key")
    # filename may be set to an empty string
    if not filename:
        filename = "private_key.pem"

    # Location may be overridden by the environment
    filename = os.environ.get("PRIVATE_KEY", filename)

    # If ctx.get_data_file() receives an absolute path, it is returned as-is
    fullpath = Path(ctx.get_data_file(filename))
    if fullpath:
        if Path(fullpath).exists():
            mounts.append(("/mnt/private_key.pem", str(fullpath)))
        else:
            warning(f"Private key file is provided {fullpath}, but does not exist")

    # Mount private keys for ssh tunnels
    ssh_tunnels = ctx.config.get("ssh-tunnels", [])
    for ssh_tunnel in ssh_tunnels:
        hostname = ssh_tunnel.get("hostname")
        key_path = ssh_tunnel.get("ssh", {}).get("identity", {}).get("key")
        if not key_path:
            error(
                f"SSH tunnel identity {Fore.RED}{hostname}{Style.RESET_ALL} "
                "key not provided. Continuing to start without this tunnel."
            )
        key_path = Path(key_path)
        if not key_path.exists():
            error(
                f"SSH tunnel identity {Fore.RED}{hostname}{Style.RESET_ALL} "
                "key does not exist. Continuing to start without this "
                "tunnel."
            )

        info(f"  Mounting private key for {hostname} at {key_path}")

        # we remove the .tmp in the container, this is because the file is
        # mounted in a volume mount point. Somehow the file is than empty in
        # the volume but not for the node instance. By removing the .tmp we
        # make sure that the file is not empty in the volume.
        mounts.append((f"/mnt/ssh/{hostname}.pem.tmp", str(key_path)))

    env = {
        "DATA_VOLUME_NAME": data_volume.name,
        "VPN_VOLUME_NAME": vpn_volume.name,
        "PRIVATE_KEY": "/mnt/private_key.pem",
    }

    # only mount the DB if it is a file
    info("Setting up databases")
    dbs = [TaskDB.from_dict(db) for db in ctx.databases]
    for db in dbs:
        # check that label contains only valid characters
        if not db.label.isidentifier():
            error(
                f"Database label {Fore.RED}{db.label}{Style.RESET_ALL} contains"
                " invalid characters. Only letters, numbers, and underscores"
                " are allowed, and it cannot start with a number."
            )
            exit(1)

        info(
            f"  Processing {Fore.GREEN}{db.type}{Style.RESET_ALL} database "
            f"{Fore.GREEN}{db.label}:{db.uri}{Style.RESET_ALL}"
        )
        label_capitals = db.label.upper()

        try:
            db_file_exists = Path(db.uri).exists()
        except Exception:
            # If the database uri cannot be parsed, it is definitely not a
            # file. In case of http servers or sql servers, checking the path
            # of the the uri will lead to an OS-dependent error, which is why
            # we catch all exceptions here.
            db_file_exists = False

        if db.type.is_file_based() and not db_file_exists:
            error(
                f"Database {Fore.RED}{db.uri}{Style.RESET_ALL} not found. Databases of "
                f"type '{db.type}' must be present on the harddrive. Please "
                "update your node configuration file."
            )
            exit(1)

        if not db_file_exists and not force_db_mount:
            debug("  - non file-based database added")
            env[f"{label_capitals}_DATABASE_URI"] = db.uri
        else:
            debug("  - file-based database added")
            suffix = Path(db.uri).suffix
            env[f"{label_capitals}_DATABASE_URI"] = f"{db.label}{suffix}"
            mounts.append((f"/mnt/{db.label}{suffix}", str(db.uri)))

    system_folders_option = "--system" if system_folders else "--user"
    cmd = (
        f"vnode-local start -c /mnt/config/{name}.yaml -n {name} "
        f" --dockerized {system_folders_option}"
    )

    volumes = []
    for mount in mounts:
        volumes.append(f"{mount[1]}:{mount[0]}")

    extra_mounts = ctx.config.get("node_extra_mounts", [])
    for mount in extra_mounts:
        volumes.append(mount)

    extra_env = ctx.config.get("node_extra_env", {})
    # all extra env var names should be valid identifiers
    extra_env_invalid = [key for key in extra_env.keys() if not key.isidentifier()]
    if extra_env_invalid:
        error(
            "Environment variable names should be valid identifiers. "
            f"The following break this rule: {extra_env_invalid}"
        )
        exit(1)
    # we won't accept overwrites of existing env vars
    env_overwrites = extra_env.keys() & env.keys()
    if env_overwrites:
        error(f"Cannot overwrite existing node environment variables: {env_overwrites}")
        exit(1)
    env.update(extra_env)

    # Add extra hosts to the environment
    extra_hosts = ctx.config.get("node_extra_hosts", {})

    remove_container_if_exists(
        docker_client=docker_client, name=ctx.docker_container_name
    )

    info("Running Docker container")
    container = docker_client.containers.run(
        image,
        command=cmd,
        volumes=volumes,
        detach=True,
        labels={
            f"{APPNAME}-type": InstanceType.NODE.value,
            "system": str(system_folders),
            "name": ctx.config_file_name,
        },
        environment=env,
        name=ctx.docker_container_name,
        auto_remove=not keep,
        tty=True,
        extra_hosts=extra_hosts,
    )

    info("Node container was started!")
    info(
        "Please check the node logs to see if the node successfully connects to the "
        "server."
    )

    if attach:
        logs = container.attach(stream=True, logs=True)
        Thread(target=print_log_worker, args=(logs,), daemon=True).start()
        while True:
            try:
                time.sleep(1)
            except KeyboardInterrupt:
                info("Closing log file. Keyboard Interrupt.")
                info(
                    "Note that your node is still running! Shut it down with "
                    f"'{Fore.RED}v6 node stop{Style.RESET_ALL}'"
                )
                exit(0)
    else:
        info(
            f"To see the logs, run: {Fore.GREEN}v6 node attach --name "
            f"{ctx.name}{Style.RESET_ALL}"
        )



from functools import wraps
from pathlib import Path

import click

from vantage6.common import error
from vantage6.common.globals import InstanceType

from vantage6.cli.configuration_wizard import select_configuration_questionaire
from vantage6.cli.context import get_context, select_context_class


def click_insert_context(
    type_: InstanceType,
    include_name: bool = False,
    include_system_folders: bool = False,
) -> callable:
    """
    Supply the Click function with an additional context parameter. The context
    is passed to the function as the first argument.

    Parameters
    ----------
    type_ : InstanceType
        The type of instance for which the context should be inserted
    include_name : bool
        Include the name of the configuration as an argument
    include_system_folders : bool
        Include whether or not to use the system folders as an argument

    Returns
    -------
    Callable
        Click function with context

    Examples
    --------
    >>> @click_insert_context(InstanceType.SERVER)
    >>> def cli_server_start(ctx: ServerContext, *args, **kwargs) -> None:
    >>>     pass
    """

    def protection_decorator(func: callable) -> callable:
        @click.option("-n", "--name", default=None, help="Name of the configuration.")
        @click.option(
            "-c",
            "--config",
            default=None,
            help="Path to configuration-file; overrides --name",
        )
        @click.option(
            "--system",
            "system_folders",
            flag_value=True,
            help="Use system folders instead of user folders. This is the default",
        )
        @click.option(
            "--user",
            "system_folders",
            flag_value=False,
            default=False if type_ == InstanceType.NODE else True,
            help="Use user folders instead of system folders",
        )
        @wraps(func)
        def decorator(
            name: str, config: str, system_folders: bool, *args, **kwargs
        ) -> callable:
            """
            Decorator function that adds the context to the function.

            Returns
            -------
            Callable
                Decorated function
            """
            ctx_class = select_context_class(type_)
            # path to configuration file always overrides name
            if config:
                ctx = ctx_class.from_external_config_file(config, system_folders)
            elif "ctx" in kwargs:
                # if ctx is already in kwargs (typically when one click command
                # calls another internally), use that existing ctx
                ctx = kwargs.pop("ctx")
            else:
                # in case no name, ctx or config file is supplied, ask the user
                # to select an existing config by name
                if not name:
                    try:
                        # select configuration if none supplied
                        name = select_configuration_questionaire(type_, system_folders)
                    except Exception:
                        error("No configurations could be found!")
                        exit(1)

                ctx = get_context(type_, name, system_folders)
            extra_args = []
            if include_name:
                if not name:
                    name = Path(config).stem
                extra_args.append(name)
            if include_system_folders:
                extra_args.append(system_folders)
            return func(ctx, *extra_args, *args, **kwargs)

        return decorator

    return protection_decorator



import click
import questionary
from kubernetes import client, config

from vantage6.cli.config import CliConfig
from vantage6.cli.utils import switch_context_and_namespace
from vantage6.common import error


@click.command()
@click.argument("namespace", required=False, type=str, default=None)
def cli_use_namespace(namespace: str):
    """
    Set which Kubernetes namespace to use.

    The namespace will be created if it does not exist.
    """
    # Load the active context configuration
    config.load_kube_config()

    try:
        v1 = client.CoreV1Api()
        namespace_list = v1.list_namespace()
    except Exception:
        error(
            "Failed to connect to Kubernetes cluster. Check if the cluster is running and reachable."
        )
        return

    namespace_names = [ns.metadata.name for ns in namespace_list.items]

    # Prompt user to select a namespace
    if not namespace:
        namespace = questionary.select(
            "Which Kubernetes namespace do you want to use?",
            choices=namespace_names,
        ).ask()

        # If no namespace is selected (e.g. KeyboardInterrupt), exit
        if not namespace:
            return

    # Check if namespace exists, create if not
    if namespace not in namespace_names:
        namespace_body = client.V1Namespace(
            metadata=client.V1ObjectMeta(name=namespace)
        )
        v1.create_namespace(namespace_body)

    # Switch to the selected namespace for current context
    switch_context_and_namespace(namespace=namespace)

    # Remove the last_context in CLI config
    cli_config = CliConfig()
    cli_config.remove_kube()



from __future__ import annotations

import re
import subprocess
import time
from os import PathLike
from pathlib import Path

import docker
from docker.client import DockerClient

from vantage6.common import error, info, warning
from vantage6.common.docker.addons import pull_image
from vantage6.common.globals import (
    DEFAULT_ALGO_STORE_IMAGE,
    DEFAULT_CHART_REPO,
    DEFAULT_DOCKER_REGISTRY,
    DEFAULT_NODE_IMAGE,
    DEFAULT_SERVER_IMAGE,
    DEFAULT_UI_IMAGE,
    InstanceType,
)

from vantage6.cli.globals import ChartName
from vantage6.cli.utils import validate_input_cmd_args


def pull_infra_image(
    client: DockerClient, image: str, instance_type: InstanceType
) -> None:
    """
    Try to pull an infrastructure image. If the image is a default infrastructure image,
    exit if in cannot be pulled. If it is not a default image, exit if it cannot be
    pulled and it is also not available locally. If a local image is available, a
    warning is printed.

    Parameters
    ----------
    client : DockerClient
        The Docker client
    image : str
        The image name to pull
    instance_type : InstanceType
        The type of instance to pull the image for
    """
    try:
        pull_image(client, image, suppress_error=True)
    except docker.errors.APIError:
        if not _is_default_infra_image(image, instance_type):
            if _image_exists_locally(client, image):
                warning("Failed to pull infrastructure image! Will use local image...")
            else:
                error("Failed to pull infrastructure image!")
                error("Image also not found locally. Exiting...")
                exit(1)
        else:
            error("Failed to pull infrastructure image! Exiting...")
            exit(1)


def _is_default_infra_image(image: str, instance_type: InstanceType) -> bool:
    """
    Check if an infrastructure image is the default image.

    Parameters
    ----------
    image : str
        The image name to check
    instance_type : InstanceType
        The type of instance to check the image for

    Returns
    -------
    bool
        True if the image is the default image, False otherwise
    """
    if instance_type == InstanceType.SERVER:
        return image == f"{DEFAULT_DOCKER_REGISTRY}/{DEFAULT_SERVER_IMAGE}"
    elif instance_type == InstanceType.ALGORITHM_STORE:
        return image == f"{DEFAULT_DOCKER_REGISTRY}/{DEFAULT_ALGO_STORE_IMAGE}"
    elif instance_type == InstanceType.UI:
        return image == f"{DEFAULT_DOCKER_REGISTRY}/{DEFAULT_UI_IMAGE}"
    elif instance_type == InstanceType.NODE:
        return image == f"{DEFAULT_DOCKER_REGISTRY}/{DEFAULT_NODE_IMAGE}"


def _image_exists_locally(client: DockerClient, image: str) -> bool:
    """
    Check if the image exists locally.

    Parameters
    ----------
    client : DockerClient
        The Docker client
    image : str
        The image name to check

    Returns
    -------
    bool
        True if the image exists locally, False otherwise
    """
    try:
        client.images.get(image)
    except docker.errors.ImageNotFound:
        return False
    return True


def helm_install(
    release_name: str,
    chart_name: ChartName,
    values_file: str | PathLike | None = None,
    context: str | None = None,
    namespace: str | None = None,
) -> None:
    """
    Manage the `helm install` command.

    Parameters
    ----------
    release_name : str
        The name of the Helm release.
    chart_name : str
        The name of the Helm chart.
    values_file : str, optional
        A single values file to use with the `-f` flag.
    context : str, optional
        The Kubernetes context to use.
    namespace : str, optional
        The Kubernetes namespace to use.
    """
    # Input validation
    validate_input_cmd_args(release_name, "release name")
    validate_input_cmd_args(chart_name, "chart name")

    values_file = Path(values_file) if values_file else None
    if values_file and not values_file.is_file():
        error(f"Helm chart values file does not exist: {values_file}")
        return

    validate_input_cmd_args(context, "context name", allow_none=True)
    validate_input_cmd_args(namespace, "namespace name", allow_none=True)

    # Create the command
    command = [
        "helm",
        "install",
        release_name,
        chart_name,
        "--repo",
        DEFAULT_CHART_REPO,
        # TODO v5+ remove this flag when we have a stable release
        "--devel",  # ensure using latest version including pre-releases
    ]

    if values_file:
        command.extend(["-f", str(values_file)])

    if context:
        command.extend(["--kube-context", context])

    if namespace:
        command.extend(["--namespace", namespace])

    try:
        subprocess.run(
            command,
            stdout=subprocess.DEVNULL,
            check=True,
        )
        info(
            f"Successfully installed release '{release_name}' using chart "
            f"'{chart_name}'."
        )
    except subprocess.CalledProcessError:
        error(f"Failed to install release '{release_name}'.")
        exit(1)
    except FileNotFoundError:
        error(
            "Helm command not found. Please ensure Helm is installed and available in "
            "the PATH."
        )
        exit(1)


def start_port_forward(
    service_name: str,
    service_port: int,
    port: int,
    ip: str | None,
    context: str | None = None,
    namespace: str | None = None,
) -> None:
    """
    Port forward a kubernetes service.

    Parameters
    ----------
    service_name : str
        The name of the Kubernetes service to port forward.
    service_port : int
        The port on the service to forward.
    port : int
        The port to listen on.
    ip : str | None
        The IP address to listen on. If None, defaults to localhost.
    context : str | None
        The Kubernetes context to use.
    namespace : str | None
        The Kubernetes namespace to use.
    """
    # Input validation
    validate_input_cmd_args(service_name, "service name")
    if not isinstance(service_port, int) or service_port <= 0:
        error(f"Invalid service port: {service_port}. Must be a positive integer.")
        return

    if not isinstance(port, int) or port <= 0:
        error(f"Invalid local port: {port}. Must be a positive integer.")
        return

    if ip and not re.match(
        r"^(localhost|[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3})$", ip
    ):
        error(f"Invalid IP address: {ip}. Must be a valid IPv4 address or 'localhost'.")
        return

    validate_input_cmd_args(context, "context name", allow_none=True)
    validate_input_cmd_args(namespace, "namespace name", allow_none=True)

    # Check if the service is ready before starting port forwarding
    info(f"Waiting for service '{service_name}' to become ready...")
    start_time = time.time()
    timeout = 300  # seconds
    while time.time() - start_time < timeout:
        try:
            result = (
                subprocess.check_output(
                    [
                        "kubectl",
                        "get",
                        "endpoints",
                        service_name,
                        "-o",
                        "jsonpath={.subsets[*].addresses[*].ip}",
                    ]
                )
                .decode()
                .strip()
            )

            if result:
                info(f"Service '{service_name}' is ready.")
                break
        except subprocess.CalledProcessError:
            pass  # ignore and retry

        time.sleep(2)
    else:
        error(
            f"Timeout: Service '{service_name}' has no ready endpoints after {timeout} seconds."
        )
        return

    # Create the port forwarding command
    if not ip:
        ip = "localhost"

    command = [
        "kubectl",
        "port-forward",
        "--address",
        ip,
        f"service/{service_name}",
        f"{port}:{service_port}",
    ]

    if context:
        command.extend(["--context", context])

    if namespace:
        command.extend(["--namespace", namespace])

    # Start the port forwarding process
    try:
        process = subprocess.Popen(
            command,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.PIPE,
            start_new_session=True,  # Start in new session to detach from parent
        )

        # Give the process a moment to start and check if it's still running
        time.sleep(1)
        if process.poll() is not None:
            # Process has already terminated
            e = process.stderr.read().decode() if process.stderr else "Unknown error"
            error(f"Failed to start port forwarding: {e}")
            return

        info(
            f"Port forwarding started: {ip}:{port} -> {service_name}:{service_port} "
            f"(PID: {str(process.pid)})"
        )
        return
    except Exception as e:
        error(f"Failed to start port forwarding: {e}")
        return



from typing import Self

from schema import And, Optional, Or, Use

from vantage6.common.configuration_manager import Configuration, ConfigurationManager

from vantage6.cli.globals import ALGO_STORE_TEMPLATE_FILE, SERVER_TEMPLATE_FILE

LOGGING_VALIDATORS = {
    "level": And(
        Use(str), lambda lvl: lvl in ("DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL")
    ),
    "use_console": Use(bool),
    "backup_count": And(Use(int), lambda n: n > 0),
    "max_size": And(Use(int), lambda b: b > 16),
    "format": Use(str),
    "datefmt": Use(str),
}


class ServerConfiguration(Configuration):
    """
    Stores the server's configuration and defines a set of server-specific
    validators.
    """

    # TODO: explore how to validate helm values.yaml files, see issue 2105
    VALIDATORS = {}


class AlgorithmStoreConfiguration(Configuration):
    """
    Stores the algorithm store's configuration and defines a set of algorithm store-specific
    validators.
    """

    VALIDATORS = {}


class NodeConfiguration(Configuration):
    """
    Stores the node's configuration and defines a set of node-specific
    validators.
    """

    VALIDATORS = {
        "server_url": Use(str),
        "port": Or(Use(int), None),
        "task_dir": Use(str),
        # TODO: remove `dict` validation from databases
        "api_path": Use(str),
        "logging": LOGGING_VALIDATORS,
        "encryption": {"enabled": bool, Optional("private_key"): Use(str)},
        Optional("node_extra_env"): dict,
        Optional("node_extra_mounts"): [str],
        Optional("node_extra_hosts"): dict,
        Optional("share_algorithm_logs"): Use(bool),
    }


class TestConfiguration(Configuration):
    VALIDATORS = {}


class NodeConfigurationManager(ConfigurationManager):
    """
    Maintains the node's configuration.

    Parameters
    ----------
    name : str
        Name of the configuration file.
    """

    def __init__(self, name, *args, **kwargs) -> None:
        super().__init__(conf_class=NodeConfiguration, name=name)

    @classmethod
    def from_file(cls, path: str) -> Self:
        """
        Create a new instance of the NodeConfigurationManager from a
        configuration file.

        Parameters
        ----------
        path : str
            Path to the configuration file.

        Returns
        -------
        NodeConfigurationManager
            A new instance of the NodeConfigurationManager.
        """
        return super().from_file(path, conf_class=NodeConfiguration)


class ServerConfigurationManager(ConfigurationManager):
    """
    Maintains the server's configuration.

    Parameters
    ----------
    name : str
        Name of the configuration file.
    """

    def __init__(self, name, *args, **kwargs) -> None:
        super().__init__(conf_class=ServerConfiguration, name=name)

    @classmethod
    def from_file(cls, path) -> Self:
        """
        Create a new instance of the ServerConfigurationManager from a
        configuration file.

        Parameters
        ----------
        path : str
            Path to the configuration file.

        Returns
        -------
        ServerConfigurationManager
            A new instance of the ServerConfigurationManager.
        """
        return super().from_file(path, conf_class=ServerConfiguration)

    def get_config_template(self) -> str:
        """
        Get the configuration template for the server.

        Returns
        -------
        str
            The configuration template for the server.
        """
        return super()._get_config_template(SERVER_TEMPLATE_FILE)


class AlgorithmStoreConfigurationManager(ConfigurationManager):
    """
    Maintains the algorithm store's configuration.

    Parameters
    ----------
    name : str
        Name of the configuration file.
    """

    def __init__(self, name, *args, **kwargs) -> None:
        super().__init__(conf_class=AlgorithmStoreConfiguration, name=name)

    @classmethod
    def from_file(cls, path: str) -> Self:
        """
        Create a new instance of the AlgorithmStoreConfigurationManager from a
        configuration file.

        Parameters
        ----------
        path : str
            Path to the configuration file.

        Returns
        -------
        AlgorithmStoreConfigurationManager
            A new instance of the AlgorithmStoreConfigurationManager.
        """
        return super().from_file(path, conf_class=AlgorithmStoreConfiguration)

    def get_config_template(self) -> str:
        """
        Get the configuration template for the algorithm store.

        Returns
        -------
        str
            The configuration template for the algorithm store.
        """
        return super()._get_config_template(ALGO_STORE_TEMPLATE_FILE)


class TestingConfigurationManager(ConfigurationManager):
    def __init__(self, name, *args, **kwargs):
        super().__init__(conf_class=TestConfiguration, name=name)

    @classmethod
    def from_file(cls, path):
        return super().from_file(path, conf_class=TestConfiguration)
